#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The fmriprep reporting engine for visual assessment
"""

from fmriprep.viz.pdf_compose import generate_report
from fmriprep.viz.pipeline_reports import (anatomical_overlay, parcel_overlay,
                               stripped_brain_overlay, generate_report_workflow)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# @Author: oesteban
# @Date:   2016-01-05 11:29:40
# @Email:  code@oscaresteban.es
# @Last modified by:   oesteban
# @Last Modified time: 2016-07-21 10:19:45
"""
Utilities for data grabbers (from nilearn)
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import os.path as op
import sys
import shutil
import time
import base64
import hashlib
import subprocess as sp
from fmriprep.data.compat import _urllib, md5_hash


def _fetch_file(url, dataset_dir, filetype=None, resume=True, overwrite=False,
                md5sum=None, username=None, password=None, handlers=None,
                verbose=1, temp_downloads=None):
    """Load requested file, downloading it if needed or requested.

    :param str url: contains the url of the file to be downloaded.
    :param str dataset_dir: path of the data directory. Used for data
        storage in the specified location.
    :param bool resume: if true, try to resume partially downloaded files
    :param overwrite: if bool true and file already exists, delete it.
    :param str md5sum: MD5 sum of the file. Checked if download of the file
        is required
    :param str username: username used for basic HTTP authentication
    :param str password: password used for basic HTTP authentication
    :param list(BaseHandler) handlers: urllib handlers passed to
        urllib.request.build_opener. Used by advanced users to customize
        request handling.
    :param int verbose: verbosity level (0 means no message).
    :returns: absolute path of downloaded file
    :rtype: str

    ..note::

      If, for any reason, the download procedure fails, all downloaded files are
      removed.


    """
    if handlers is None:
        handlers = []

    if not overwrite and os.listdir(dataset_dir):
        return True

    data_dir, _ = op.split(dataset_dir)

    if temp_downloads is None:
        temp_downloads = op.join(op.expanduser('~/.cache/stanford-crn'), 'downloads')

    # Determine data path
    if not op.exists(temp_downloads):
        os.makedirs(temp_downloads)

    # Determine filename using URL
    parse = _urllib.parse.urlparse(url)
    file_name = op.basename(parse.path)
    if file_name == '':
        file_name = md5_hash(parse.path)

        if filetype is not None:
            file_name += filetype

    temp_full_name = op.join(temp_downloads, file_name)
    temp_part_name = temp_full_name + ".part"

    if overwrite:
        shutil.rmtree(dataset_dir, ignore_errors=True)

    if op.exists(temp_full_name):
        if overwrite:
            os.remove(temp_full_name)

    t_0 = time.time()
    local_file = None
    initial_size = 0

    try:
        # Download data
        url_opener = _urllib.request.build_opener(*handlers)
        request = _urllib.request.Request(url)
        request.add_header('Connection', 'Keep-Alive')
        if username is not None and password is not None:
            if not url.startswith('https'):
                raise ValueError(
                    'Authentication was requested on a non  secured URL (%s).'
                    'Request has been blocked for security reasons.' % url)
            # Note: HTTPBasicAuthHandler is not fitted here because it relies
            # on the fact that the server will return a 401 error with proper
            # www-authentication header, which is not the case of most
            # servers.
            encoded_auth = base64.b64encode(
                (username + ':' + password).encode())
            request.add_header(b'Authorization', b'Basic ' + encoded_auth)
        if verbose > 0:
            displayed_url = url.split('?')[0] if verbose == 1 else url
            print('Downloading data from %s ...' % displayed_url)
        if resume and op.exists(temp_part_name):
            # Download has been interrupted, we try to resume it.
            local_file_size = op.getsize(temp_part_name)
            # If the file exists, then only download the remainder
            request.add_header("Range", "bytes=%s-" % (local_file_size))
            try:
                data = url_opener.open(request)
                content_range = data.info().get('Content-Range')
                if (content_range is None or not content_range.startswith(
                        'bytes %s-' % local_file_size)):
                    raise IOError('Server does not support resuming')
            except Exception:
                # A wide number of errors can be raised here. HTTPError,
                # URLError... I prefer to catch them all and rerun without
                # resuming.
                if verbose > 0:
                    print('Resuming failed, try to download the whole file.')
                return _fetch_file(
                    url, dataset_dir, resume=False, overwrite=overwrite,
                    md5sum=md5sum, username=username, password=password,
                    handlers=handlers, verbose=verbose)
            local_file = open(temp_part_name, "ab")
            initial_size = local_file_size
        else:
            data = url_opener.open(request)
            local_file = open(temp_part_name, "wb")
        _chunk_read_(data, local_file, report_hook=(verbose > 0),
                     initial_size=initial_size, verbose=verbose)
        # temp file must be closed prior to the move
        if not local_file.closed:
            local_file.close()
        shutil.move(temp_part_name, temp_full_name)
        delta_t = time.time() - t_0
        if verbose > 0:
            # Complete the reporting hook
            sys.stderr.write(' ...done. ({0:.0f} seconds, {1:.0f} min)\n'
                             .format(delta_t, delta_t // 60))
    except (_urllib.error.HTTPError, _urllib.error.URLError) as exc:
        if 'Error while fetching' not in str(exc):
            # For some odd reason, the error message gets doubled up
            #   (possibly from the re-raise), so only add extra info
            #   if it's not already there.
            exc.reason = ("%s| Error while fetching file %s; "
                          "dataset fetching aborted." % (
                              str(exc.reason), file_name))
        raise
    finally:
        if local_file is not None:
            if not local_file.closed:
                local_file.close()

    if md5sum is not None:
        if _md5_sum_file(temp_full_name) != md5sum:
            raise ValueError("File %s checksum verification has failed."
                             " Dataset fetching aborted." % local_file)

    if filetype is None:
        fname, filetype = op.splitext(op.basename(temp_full_name))
        if filetype == '.gz':
            fname, ext = op.splitext(fname)
            filetype = ext + filetype

    if filetype.startswith('.'):
        filetype = filetype[1:]

    if filetype == 'tar':
        sp.check_call(['tar', 'xf', temp_full_name], cwd=data_dir)
        os.remove(temp_full_name)
        return True

    return True

def _get_dataset_dir(dataset_name, data_dir=None, default_paths=None,
                     verbose=1):
    """ Create if necessary and returns data directory of given dataset.

    :param str dataset_name: The unique name of the dataset.
    :param str data_dir: Path of the data directory. Used to force data storage in
      a specified location.
    :param list(str) default_paths: Default system paths in which the dataset
      may already have been installed by a third party software. They will be
      checked first.
    :param int verbose: verbosity level (0 means no message).

    :returns: the path of the given dataset directory.
    :rtype: str

    .. note::

      This function retrieves the datasets directory (or data directory) using
      the following priority :

        1. defaults system paths
        2. the keyword argument data_dir
        3. the global environment variable CRN_SHARED_DATA
        4. the user environment variable CRN_DATA
        5. ~/.cache/stanford-crn in the user home folder


    """
    # We build an array of successive paths by priority
    # The boolean indicates if it is a pre_dir: in that case, we won't add the
    # dataset name to the path.
    paths = []

    # Check data_dir which force storage in a specific location
    if data_dir is not None:
        paths.extend([(d, False) for d in data_dir.split(os.pathsep)])

    # Search possible system paths
    if default_paths is not None:
        for default_path in default_paths:
            paths.extend([(d, True) for d in default_path.split(os.pathsep)])

    # If data_dir has not been specified, then we crawl default locations
    if data_dir is None:
        global_data = os.getenv('CRN_SHARED_DATA')
        if global_data is not None:
            paths.extend([(d, False) for d in global_data.split(os.pathsep)])

        local_data = os.getenv('CRN_DATA')
        if local_data is not None:
            paths.extend([(d, False) for d in local_data.split(os.pathsep)])

        paths.append((op.expanduser('~/.cache/stanford-crn'), False))

    if verbose > 2:
        print('Dataset search paths: %s' % paths)

    # Check if the dataset exists somewhere
    for path, is_pre_dir in paths:
        if not is_pre_dir:
            path = op.join(path, dataset_name)
        if op.islink(path):
            # Resolve path
            path = readlinkabs(path)
        if op.exists(path) and op.isdir(path):
            if verbose > 1:
                print('\nDataset found in %s\n' % path)
            return path

    # If not, create a folder in the first writeable directory
    errors = []
    for (path, is_pre_dir) in paths:
        if not is_pre_dir:
            path = op.join(path, dataset_name)
        if not op.exists(path):
            try:
                os.makedirs(path)
                if verbose > 0:
                    print('\nDataset created in %s\n' % path)
                return path
            except Exception as exc:
                short_error_message = getattr(exc, 'strerror', str(exc))
                errors.append('\n -{0} ({1})'.format(
                    path, short_error_message))

    raise OSError('fmriprep tried to store the dataset in the following '
                  'directories, but:' + ''.join(errors))


def readlinkabs(link):
    """
    Return an absolute path for the destination
    of a symlink
    """
    path = os.readlink(link)
    if op.isabs(path):
        return path
    return op.join(op.dirname(link), path)


def _md5_sum_file(path):
    """ Calculates the MD5 sum of a file.
    """
    with open(path, 'rb') as fhandle:
        md5sum = hashlib.md5()
        while True:
            data = fhandle.read(8192)
            if not data:
                break
            md5sum.update(data)
    return md5sum.hexdigest()


def _chunk_read_(response, local_file, chunk_size=8192, report_hook=None,
                 initial_size=0, total_size=None, verbose=1):
    """Download a file chunk by chunk and show advancement

    :param _urllib.response.addinfourl response: response to the download
        request in order to get file size
    :param str local_file: hard disk file where data should be written
    :param int chunk_size: size of downloaded chunks. Default: 8192
    :param bool report_hook: whether or not to show downloading advancement
    :param int initial_size: if resuming, indicate the initial size of the file
    :param int total_size: Expected final size of download (None means it
        is unknown).
    :param int verbose: verbosity level (0 means no message).
    :returns: the downloaded file path.
    :rtype: string

    """
    try:
        if total_size is None:
            total_size = response.info().get('Content-Length').strip()
        total_size = int(total_size) + initial_size
    except Exception as exc:
        if verbose > 2:
            print("Warning: total size could not be determined.")
            if verbose > 3:
                print("Full stack trace: %s" % exc)
        total_size = None
    bytes_so_far = initial_size

    t_0 = time_last_display = time.time()
    while True:
        chunk = response.read(chunk_size)
        bytes_so_far += len(chunk)
        time_last_read = time.time()
        if (report_hook and
                # Refresh report every half second or when download is
                # finished.
                (time_last_read > time_last_display + 0.5 or not chunk)):
            _chunk_report_(bytes_so_far,
                           total_size, initial_size, t_0)
            time_last_display = time_last_read
        if chunk:
            local_file.write(chunk)
        else:
            break

    return

def _chunk_report_(bytes_so_far, total_size, initial_size, t_0):
    """Show downloading percentage.

    :param int bytes_so_far: number of downloaded bytes
    :param int total_size: total size of the file (may be 0/None, depending
        on download method).
    :param int t_0: the time in seconds (as returned by time.time()) at which
        the download was resumed / started.
    :param int initial_size: if resuming, indicate the initial size of the
        file. If not resuming, set to zero.
    """

    if not total_size:
        sys.stderr.write("\rDownloaded %d of ? bytes." % (bytes_so_far))

    else:
        # Estimate remaining download time
        total_percent = float(bytes_so_far) / total_size

        current_download_size = bytes_so_far - initial_size
        bytes_remaining = total_size - bytes_so_far
        delta_t = time.time() - t_0
        download_rate = current_download_size / max(1e-8, float(delta_t))
        # Minimum rate of 0.01 bytes/s, to avoid dividing by zero.
        time_remaining = bytes_remaining / max(0.01, download_rate)

        # Trailing whitespace is to erase extra char when message length
        # varies
        sys.stderr.write(
            "\rDownloaded %d of %d bytes (%.1f%%, %s remaining)"
            % (bytes_so_far, total_size, total_percent * 100,
               _format_time(time_remaining)))


def _format_time(t_secs):
    if t_secs > 60:
        return "%4.1fmin" % (t_secs / 60.)
    else:
        return " %5.1fs" % (t_secs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import os.path as op

from nipype.interfaces import fsl
from nipype.interfaces import io as nio
from nipype.interfaces import utility as niu
from nipype.interfaces.ants.segmentation import N4BiasFieldCorrection
from nipype.pipeline import engine as pe

from fmriprep.utils.misc import gen_list
from fmriprep.interfaces import ReadSidecarJSON
from fmriprep.viz import stripped_brain_overlay

WORKFLOW_NAME = 'Fieldmap_SEs'

# pylint: disable=R0914
def se_pair_workflow(name=WORKFLOW_NAME, settings=None):
    """
    Estimates the fieldmap using TOPUP on series of :abbr:`SE (Spin-Echo)` images
    acquired with varying :abbr:`PE (phase encoding)` direction.
    """

    if settings is None:
        settings = {}

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=['fieldmaps']), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=[
        'out_field', 'fmap_mask', 'mag_brain', 'fmap_fieldcoef', 'fmap_movpar']), name='outputnode')

    # Read metadata
    meta = pe.MapNode(ReadSidecarJSON(fields=['TotalReadoutTime', 'PhaseEncodingDirection']),
                      iterfield=['in_file'], name='metadata')

    encfile = pe.Node(interface=niu.Function(
        input_names=['fieldmaps', 'in_dict'], output_names=['parameters_file'],
        function=create_encoding_file), name='TopUp_encfile', updatehash=True)

    # Head motion correction
    fslmerge = pe.Node(fsl.Merge(dimension='t'), name='SE_merge')
    hmc_se = pe.Node(fsl.MCFLIRT(cost='normcorr', mean_vol=True), name='SE_head_motion_corr')
    fslsplit = pe.Node(fsl.Split(dimension='t'), name='SE_split')

    # Run topup to estimate field distortions, do not estimate movement
    # since it is done in hmc_se
    topup = pe.Node(fsl.TOPUP(estmov=0), name='TopUp')

    # Use the least-squares method to correct the dropout of the SE images
    unwarp_mag = pe.Node(fsl.ApplyTOPUP(method='lsr'), name='TopUpApply')

    # Remove bias
    inu_n4 = pe.Node(N4BiasFieldCorrection(dimension=3), name='SE_bias')

    # Skull strip corrected SE image to get reference brain and mask
    mag_bet = pe.Node(fsl.BET(mask=True, robust=True), name='SE_brain')

    workflow.connect([
        (inputnode, meta, [('fieldmaps', 'in_file')]),
        (inputnode, encfile, [('fieldmaps', 'fieldmaps')]),
        (inputnode, fslmerge, [('fieldmaps', 'in_files')]),
        (fslmerge, hmc_se, [('merged_file', 'in_file')]),
        (meta, encfile, [('out_dict', 'in_dict')]),
        (encfile, topup, [('parameters_file', 'encoding_file')]),
        (hmc_se, topup, [('out_file', 'in_file')]),
        (topup, unwarp_mag, [('out_fieldcoef', 'in_topup_fieldcoef'),
                             ('out_movpar', 'in_topup_movpar')]),
        (encfile, unwarp_mag, [('parameters_file', 'encoding_file')]),
        (hmc_se, fslsplit, [('out_file', 'in_file')]),
        (fslsplit, unwarp_mag, [('out_files', 'in_files'),
                                (('out_files', gen_list), 'in_index')]),
        (unwarp_mag, inu_n4, [('out_corrected', 'input_image')]),
        (inu_n4, mag_bet, [('output_image', 'in_file')]),

        (topup, outputnode, [('out_field', 'out_field')]),
        (mag_bet, outputnode, [('out_file', 'mag_brain'),
                               ('mask_file', 'fmap_mask')]),
        (topup, outputnode, [('out_fieldcoef', 'fmap_fieldcoef'),
                             ('out_movpar', 'fmap_movpar')])
    ])

    # Reports section
    se_png = pe.Node(niu.Function(
        input_names=['in_file', 'overlay_file', 'out_file'], output_names=['out_file'],
        function=stripped_brain_overlay), name='PNG_SE_corr')
    se_png.inputs.out_file = 'corrected_SE_and_mask.png'

    datasink = pe.Node(nio.DataSink(base_directory=op.join(settings['work_dir'], 'images')),
                       name='datasink', parameterization=False)
    workflow.connect([
        (unwarp_mag, se_png, [('out_corrected', 'overlay_file')]),
        (mag_bet, se_png, [('mask_file', 'in_file')]),
        (se_png, datasink, [('out_file', '@corrected_SE_and_mask')])
    ])

    return workflow

def create_encoding_file(fieldmaps, in_dict):
    """Creates a valid encoding file for topup"""
    import json
    import nibabel as nb
    import numpy as np
    import os

    if not isinstance(fieldmaps, list):
        fieldmaps = [fieldmaps]
    if not isinstance(in_dict, list):
        in_dict = [in_dict]

    pe_dirs = {'i': 0, 'j': 1, 'k': 2}
    enc_table = []
    for fmap, meta in zip(fieldmaps, in_dict):
        line_values = [0, 0, 0, meta['TotalReadoutTime']]
        line_values[pe_dirs[meta['PhaseEncodingDirection'][0]]] = 1 + (
            -2*(len(meta['PhaseEncodingDirection']) == 2))

        nvols = 1
        if len(nb.load(fmap).shape) > 3:
            nvols = nb.load(fmap).shape[3]

        enc_table += [line_values] * nvols

    np.savetxt(os.path.abspath('parameters.txt'), enc_table, fmt=['%0.1f', '%0.1f', '%0.1f', '%0.20f'])
    return os.path.abspath('parameters.txt')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Created on Wed Dec  2 17:35:40 2015

@author: craigmoodie
"""

from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype.interfaces import fsl
from nipype.interfaces import freesurfer as fs
import nipype.interfaces.io as nio

from fmriprep.workflows.anatomical import t1w_preprocessing
from fmriprep.workflows.fieldmap.se_pair_workflow import se_pair_workflow
from fmriprep.workflows.fieldmap.fieldmap_to_phasediff import fieldmap_to_phasediff
from fmriprep.workflows.fieldmap.decider import fieldmap_decider
from fmriprep.workflows.sbref import sbref_workflow
from fmriprep.workflows import sbref
from fmriprep.workflows.epi import (epi_unwarp, epi_hmc,
    epi_mean_t1_registration, epi_mni_transformation)



def fmri_preprocess_single(subject_data, name='fMRI_prep', settings=None):
    """
    The main fmri preprocessing workflow.
    """

    if settings is None:
        settings = {}

    for key in ['fsl', 'skull_strip', 'epi', 'connectivity']:
        if settings.get(key) is None:
            settings[key] = {}

    sbref_present = len(subject_data['sbref']) > 0

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(
        fields=['fieldmaps', 'fieldmaps_meta', 'epi_meta', 'sbref',
                'sbref_meta', 't1']), name='inputnode')

    # Set inputs: epi is iterable over the available runs
    for key in subject_data.keys():
        if key != 'epi':
            setattr(inputnode.inputs, key, subject_data[key])

    inputfmri = pe.Node(niu.IdentityInterface(
        fields=['epi']), name='inputfmri')
    inputfmri.iterables = [('epi', subject_data['epi'])]


    outputnode = pe.Node(niu.IdentityInterface(
        fields=['fieldmap', 'corrected_sbref', 'fmap_mag', 'fmap_mag_brain',
                't1', 'stripped_epi', 'corrected_epi_mean', 'sbref_brain',
                'stripped_epi_mask', 'stripped_t1', 't1_segmentation',
                't1_2_mni', 't1_wm_seg']),
        name='outputnode'
    )

    try:
        fmap_wf = fieldmap_decider(subject_data, settings)
    except NotImplementedError:
        fmap_wf = None

    # Reorient EPI to RAS
    split = pe.Node(fsl.Split(dimension='t'), name='SplitEPI')
    orient = pe.MapNode(fs.MRIConvert(out_type='niigz', out_orientation='RAS'),
                        iterfield=['in_file'], name='ReorientEPI')
    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeEPI')

    t1w_preproc = t1w_preprocessing(settings=settings)
    epi_hmc_wf = epi_hmc(settings=settings, sbref_present=sbref_present)
    epi_mni_trans_wf = epi_mni_transformation(settings=settings)

    #  Connecting Workflow pe.Nodes
    workflow.connect([
        (inputfmri, split, [('epi', 'in_file')]),
        (split, orient, [('out_files', 'in_file')]),
        (orient, merge, [('out_file', 'in_files')]),

        (inputnode, t1w_preproc, [('t1', 'inputnode.t1')]),
        (merge, epi_hmc_wf, [('merged_file', 'inputnode.epi_ras')]),
        (merge, epi_mni_trans_wf, [('merged_file', 'inputnode.epi_ras')]),

        # These are necessary sources for the DerivativesDataSink
        (inputfmri, epi_hmc_wf, [('epi', 'inputnode.epi')]),
        (inputfmri, epi_mni_trans_wf, [('epi', 'inputnode.epi')])
    ])

    if not sbref_present:
        epi_2_t1 = epi_mean_t1_registration(settings=settings)
        workflow.connect([
            (inputfmri, epi_2_t1, [('epi', 'inputnode.epi')]),
            (epi_hmc_wf, epi_2_t1, [('outputnode.epi_mean', 'inputnode.epi_mean')]),
            (t1w_preproc, epi_2_t1, [('outputnode.t1_brain', 'inputnode.t1_brain'),
                                     ('outputnode.t1_seg', 'inputnode.t1_seg')]),
            (epi_2_t1, epi_mni_trans_wf, [('outputnode.mat_epi_to_t1', 'inputnode.mat_epi_to_t1')]),

            (epi_hmc_wf, epi_mni_trans_wf, [('outputnode.xforms', 'inputnode.hmc_xforms'),
                                            ('outputnode.epi_mask', 'inputnode.epi_mask')]),
            (t1w_preproc, epi_mni_trans_wf, [('outputnode.t1_brain', 'inputnode.t1'),
                                             ('outputnode.t1_2_mni_forward_transform',
                                              'inputnode.t1_2_mni_forward_transform')])
        ])

    if fmap_wf:
        sbref_wf = sbref_workflow(settings=settings)
        sbref_t1 = sbref.sbref_t1_registration(settings=settings)
        unwarp_wf = epi_unwarp(settings=settings)
        sepair_wf = se_pair_workflow(settings=settings)
        workflow.connect([
            (inputnode, sepair_wf, [('fieldmaps', 'inputnode.fieldmaps')]),
            (inputnode, sbref_wf, [('sbref', 'inputnode.sbref')]),
            (inputnode, unwarp_wf, [('epi', 'inputnode.epi')]),
            (sepair_wf, sbref_wf, [
                ('outputnode.mag_brain', 'inputnode.fmap_ref_brain'),
                ('outputnode.fmap_mask', 'inputnode.fmap_mask'),
                ('outputnode.fmap_fieldcoef', 'inputnode.fmap_fieldcoef'),
                ('outputnode.fmap_movpar', 'inputnode.fmap_movpar')
            ]),
            (sbref_wf, sbref_t1, [
                ('outputnode.sbref_unwarped', 'inputnode.sbref_brain')]),
            (t1w_preproc, sbref_t1, [
                ('outputnode.t1_brain', 'inputnode.t1_brain'),
                ('outputnode.t1_seg', 'inputnode.t1_seg')]),
            (sbref_wf, unwarp_wf, [
                ('outputnode.sbref_unwarped', 'inputnode.sbref_brain')]),
            (sbref_wf, epi_hmc_wf, [
                ('outputnode.sbref_unwarped', 'inputnode.sbref_brain')]),
            (epi_hmc_wf, unwarp_wf, [
                ('outputnode.epi_brain', 'inputnode.epi_brain')]),
            (sepair_wf, unwarp_wf, [
                ('outputnode.fmap_mask', 'inputnode.fmap_mask'),
                ('outputnode.fmap_fieldcoef', 'inputnode.fmap_fieldcoef'),
                ('outputnode.fmap_movpar', 'inputnode.fmap_movpar')
            ]),
        ])

    return workflow



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# @Author: oesteban
# @Date:   2016-01-05 11:29:40
# @Email:  code@oscaresteban.es
# @Last modified by:   oesteban
# @Last Modified time: 2016-07-21 10:19:33
"""
Data grabbers
"""

from fmriprep.data.utils import _get_dataset_dir, _fetch_file

def get_mni_template_ras(data_dir=None, url=None, resume=True, verbose=1):
    """Download and load the necessary files from the mni template


    :param str data_dir: path of the data directory. Used to force data storage
        in a non-standard location.
    :param str url: download URL of the dataset. Overwrite the default URL.

    """
    if url is None:
        url = "http://googledrive.com/host/0BxI12kyv2olZemt4STY1aG80cXc"

    dataset_name = 'mni_template_RAS'
    data_dir = _get_dataset_dir(dataset_name, data_dir=data_dir, verbose=verbose)

    if _fetch_file(url, data_dir, filetype='tar', resume=resume, verbose=verbose,
                   md5sum='a4669f0e7acceae148bb39450b2b21b4'):
        return data_dir
    else:
        return None

def get_mni_template(data_dir=None, url=None, resume=True, verbose=1):
    """Download and load the necessary files from the mni template


    :param str data_dir: path of the data directory. Used to force data storage
        in a non-standard location.
    :param str url: download URL of the dataset. Overwrite the default URL.

    """
    if url is None:
        url = "http://googledrive.com/host/0BxI12kyv2olZdzRDUnBPYWZGZk0"

    dataset_name = 'mni_template'
    data_dir = _get_dataset_dir(dataset_name, data_dir=data_dir, verbose=verbose)

    if _fetch_file(url, data_dir, filetype='tar', resume=resume, verbose=verbose,
                   md5sum='debfa882b8c301cd6d75dd769e73f727'):
        return data_dir
    else:
        return None

def get_ants_oasis_template(data_dir=None, url=None, resume=True, verbose=1):
    """Download and load the necessary files from the ANTs template of the OASIS dataset.


    :param str data_dir: path of the data directory. Used to force data storage
        in a non-standard location.
    :param str url: download URL of the dataset. Overwrite the default URL.

    """
    if url is None:
        url = "http://googledrive.com/host/0BxI12kyv2olZUXRWNU9aTlRvUkk"

    dataset_name = 'ants_oasis_template'
    data_dir = _get_dataset_dir(dataset_name, data_dir=data_dir, verbose=verbose)

    if _fetch_file(url, data_dir, filetype='tar', resume=resume, verbose=verbose,
                   md5sum='34d39070b541c416333cc8b6c2fe993c'):
        return data_dir
    else:
        return None

def get_ants_oasis_template_ras(data_dir=None, url=None, resume=True, verbose=1):
    """Download and load the necessary files from the ANTs template of the OASIS dataset.


    :param str data_dir: path of the data directory. Used to force data storage
        in a non-standard location.
    :param str url: download URL of the dataset. Overwrite the default URL.

    """
    if url is None:
        url = "http://googledrive.com/host/0BxI12kyv2olZTk5VZFZHT2JlNTg"

    dataset_name = 'ants_oasis_template_ras'
    data_dir = _get_dataset_dir(dataset_name, data_dir=data_dir, verbose=verbose)

    if _fetch_file(url, data_dir, filetype='tar', resume=resume, verbose=verbose,
                   md5sum='74b2f126d59ddc8a55d76cd5af4774f7'):
        return data_dir
    else:
        return None



"""
Compatibility layer for Python 3/Python 2 single codebase
"""
import sys
import hashlib


if sys.version_info[0] == 3:
    import pickle
    import io
    import urllib

    _basestring = str
    cPickle = pickle
    StringIO = io.StringIO
    BytesIO = io.BytesIO
    _urllib = urllib
    izip = zip

    def md5_hash(string):
        m = hashlib.md5()
        m.update(string.encode('utf-8'))
        return m.hexdigest()
else:
    import cPickle
    import StringIO
    import urllib
    import urllib2
    import urlparse
    import types
    import itertools

    _basestring = basestring
    cPickle = cPickle
    StringIO = BytesIO = StringIO.StringIO
    izip = itertools.izip

    class _module_lookup(object):
        modules = [urlparse, urllib2, urllib]

        def __getattr__(self, name):
            for module in self.modules:
                if hasattr(module, name):
                    attr = getattr(module, name)
                    if not isinstance(attr, types.ModuleType):
                        return attr
            raise NotImplementedError(
                'This function has not been imported properly')

    module_lookup = _module_lookup()

    class _urllib():
        request = module_lookup
        error = module_lookup
        parse = module_lookup

    def md5_hash(string):
        m = hashlib.md5()
        m.update(string)
        return m.hexdigest()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
EPI MRI -processing workflows.

Originally coded by Craig Moodie. Refactored by the CRN Developers.
"""

import os
import os.path as op

from nipype.pipeline import engine as pe
from nipype.interfaces import ants
from nipype.interfaces import c3
from nipype.interfaces import io as nio
from nipype.interfaces import utility as niu
from nipype.interfaces import fsl
from nipype.interfaces import freesurfer as fs

from fmriprep.data import get_mni_template_ras
from fmriprep.interfaces import ReadSidecarJSON, DerivativesDataSink
from fmriprep.viz import stripped_brain_overlay
from fmriprep.workflows.fieldmap.se_pair_workflow import create_encoding_file
from fmriprep.workflows.sbref import _extract_wm


# pylint: disable=R0914
def epi_hmc(name='EPIHeadMotionCorrectionWorkflow', sbref_present=False, settings=None):
    """
    Performs :abbr:`HMC (head motion correction)` over the input
    :abbr:`EPI (echo-planar imaging)` image.
    """
    workflow = pe.Workflow(name=name)

    infields = ['epi', 'epi_ras', 't1_brain']
    if sbref_present:
        infields += ['sbref_brain']

    inputnode = pe.Node(niu.IdentityInterface(fields=infields),
                        name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(
        fields=['epi_brain', 'xforms', 'epi_mask', 'epi_mean']), name='outputnode')

    bet = pe.Node(fsl.BET(functional=True, frac=0.6), name='EPI_bet')

    # Head motion correction (hmc)
    hmc = pe.Node(fsl.MCFLIRT(save_mats=True), name='EPI_hmc')

    pick_1st = pe.Node(fsl.ExtractROI(t_min=0, t_size=1), name='EPIPickFirst')
    hcm2itk = pe.MapNode(c3.C3dAffineTool(fsl2ras=True, itk_transform=True),
                         iterfield=['transform_file'], name='hcm2itk')

    workflow.connect([
        (inputnode, pick_1st, [('epi_ras', 'in_file')]),
        (inputnode, bet, [('epi_ras', 'in_file')]),
        (bet, hmc, [('out_file', 'in_file')]),

        (hmc, hcm2itk, [('mat_file', 'transform_file')]),
        (pick_1st, hcm2itk, [('roi_file', 'source_file'),
                             ('roi_file', 'reference_file')]),
        (hcm2itk, outputnode, [('itk_transform', 'xforms')]),
        (hmc, outputnode, [('out_file', 'epi_brain')])
    ])

    # Calculate EPI mask on the average after HMC
    epi_mean = pe.Node(fsl.MeanImage(dimension='T'), name='EPI_hmc_mean')
    bet_hmc = pe.Node(fsl.BET(mask=True, frac=0.6), name='EPI_hmc_bet')
    workflow.connect([
        (hmc, epi_mean, [('out_file', 'in_file')]),
        (epi_mean, bet_hmc, [('out_file', 'in_file')]),
        (bet_hmc, outputnode, [('mask_file', 'epi_mask'),
                               ('out_file', 'epi_mean')])
    ])

    # If we have an SBRef, it should be the reference,
    # align to mean volume otherwise
    if sbref_present:
        workflow.connect([
            (inputnode, hmc, [('sbref_brain', 'ref_file')]),
        ])
    else:
        hmc.inputs.mean_vol = True

    # Write corrected file in the designated output dir
    ds_hmc = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='hmc'), name='DerivativesHMC')
    ds_mats = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='hmc'), name='DerivativesHMCmats')
    ds_mask = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='hmc_bmask'), name='DerivativesEPImask')

    workflow.connect([
        (inputnode, ds_hmc, [('epi', 'source_file')]),
        (inputnode, ds_mats, [('epi', 'source_file')]),
        (inputnode, ds_mask, [('epi', 'source_file')]),
        (hmc, ds_hmc, [('out_file', 'in_file')]),
        (hcm2itk, ds_mats, [('itk_transform', 'in_file')]),
        (bet_hmc, ds_mask, [('mask_file', 'in_file')]),
    ])

    return workflow

def epi_mean_t1_registration(name='EPIMeanNormalization', settings=None):
    """
    Uses FSL FLIRT with the BBR cost function to find the transform that
    maps the EPI space into the T1-space
    """
    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(
        fields=['epi', 'epi_mean', 't1_brain', 't1_seg']), name='inputnode'
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['mat_epi_to_t1']),
        name='outputnode'
    )

    # Extract wm mask from segmentation
    wm_mask = pe.Node(
        niu.Function(input_names=['in_file'], output_names=['out_file'],
        function=_extract_wm),
        name='WM_mask'
    )


    flt_bbr_init = pe.Node(fsl.FLIRT(dof=6, out_matrix_file='init.mat'),
        name='Flirt_BBR_init')
    flt_bbr = pe.Node(fsl.FLIRT(dof=6, cost_func='bbr'), name='Flirt_BBR')
    flt_bbr.inputs.schedule = settings['fsl'].get(
        'flirt_bbr', op.join(os.getenv('FSLDIR'), 'etc/flirtsch/bbr.sch'))

    # make equivalent warp fields
    invt_bbr = pe.Node(fsl.ConvertXFM(invert_xfm=True), name='Flirt_BBR_Inv')

    #  EPI to T1 transform matrix is from fsl, using c3 tools to convert to
    #  something ANTs will like.
    fsl2itk_fwd = pe.Node(c3.C3dAffineTool(fsl2ras=True, itk_transform=True),
                          name='fsl2itk_fwd')
    fsl2itk_inv = pe.Node(c3.C3dAffineTool(fsl2ras=True, itk_transform=True),
                          name='fsl2itk_inv')

    # Write EPI mean in T1w space
    ds_t1w = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='hmc_t1'), name='DerivHMC_T1w')
    # Write registrated file in the designated output dir
    ds_tfm_fwd = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='epi2t1w_affine'), name='DerivEPI_to_T1w_fwd')
    ds_tfm_inv = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='t1w2epi_affine'), name='DerivEPI_to_T1w_inv')

    workflow.connect([
        (inputnode, wm_mask, [('t1_seg', 'in_file')]),
        (inputnode, flt_bbr_init, [('t1_brain', 'reference')]),
        (inputnode, fsl2itk_fwd, [('t1_brain', 'reference_file'),
                                  ('epi_mean', 'source_file')]),
        (inputnode, fsl2itk_inv, [('epi_mean', 'reference_file'),
                                  ('t1_brain', 'source_file')]),
        (inputnode, flt_bbr_init, [('epi_mean', 'in_file')]),
        (flt_bbr_init, flt_bbr, [('out_matrix_file', 'in_matrix_file')]),
        (inputnode, flt_bbr, [('t1_brain', 'reference')]),
        (inputnode, flt_bbr, [('epi_mean', 'in_file')]),
        (wm_mask, flt_bbr, [('out_file', 'wm_seg')]),
        (flt_bbr, invt_bbr, [('out_matrix_file', 'in_file')]),
        (flt_bbr, fsl2itk_fwd, [('out_matrix_file', 'transform_file')]),
        (invt_bbr, fsl2itk_inv, [('out_file', 'transform_file')]),
        (fsl2itk_fwd, outputnode, [('itk_transform', 'mat_epi_to_t1')]),
        (fsl2itk_inv, outputnode, [('itk_transform', 'mat_t1_to_epi')]),
        (inputnode, ds_tfm_fwd, [('epi', 'source_file')]),
        (inputnode, ds_tfm_inv, [('epi', 'source_file')]),
        (inputnode, ds_t1w, [('epi', 'source_file')]),
        (fsl2itk_fwd, ds_tfm_fwd, [('itk_transform', 'in_file')]),
        (fsl2itk_inv, ds_tfm_inv, [('itk_transform', 'in_file')]),
        (flt_bbr, ds_t1w, [('out_file', 'in_file')])
    ])

    # Plots for report
    png_sbref_t1 = pe.Node(niu.Function(
        input_names=['in_file', 'overlay_file', 'out_file'],
        output_names=['out_file'],
        function=stripped_brain_overlay),
        name='PNG_sbref_t1'
    )
    png_sbref_t1.inputs.out_file = 'sbref_to_t1.png'

    # Write corrected file in the designated output dir
    ds_png = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='epi_to_t1'), name='DerivativesPNG')

    workflow.connect([
        (flt_bbr, png_sbref_t1, [('out_file', 'overlay_file')]),
        (inputnode, png_sbref_t1, [('t1_seg', 'in_file')]),
        (inputnode, ds_png, [('epi', 'source_file')]),
        (png_sbref_t1, ds_png, [('out_file', 'in_file')])
    ])

    return workflow

def epi_mni_transformation(name='EPIMNITransformation', settings=None):
    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'mat_epi_to_t1',
            't1_2_mni_forward_transform',
            'epi',
            'epi_ras',
            'epi_mask',
            't1',
            'hmc_xforms'
        ]),
        name='inputnode'
    )

    def _aslist(in_value):
        if isinstance(in_value, list):
            return in_value
        return [in_value]

    pick_1st = pe.Node(fsl.ExtractROI(t_min=0, t_size=1), name='EPIPickFirst')

    gen_ref = pe.Node(niu.Function(
        input_names=['fixed_image', 'moving_image'], output_names=['out_file'],
        function=_gen_reference), name='GenNewMNIReference')
    gen_ref.inputs.fixed_image = op.join(get_mni_template_ras(),
                                         'MNI152_T1_1mm.nii.gz')

    split = pe.Node(fsl.Split(dimension='t'), name='SplitEPI')
    merge_transforms = pe.MapNode(niu.Merge(3),
                                  iterfield=['in3'], name='MergeTransforms')
    epi_to_mni_transform = pe.MapNode(
        ants.ApplyTransforms(), iterfield=['input_image', 'transforms'],
        name='EPIToMNITransform')
    epi_to_mni_transform.terminal_output = 'file'
    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeEPI')

    mask_merge_tfms = pe.Node(niu.Merge(2), name='MaskMergeTfms')
    mask_mni_tfm = pe.Node(ants.ApplyTransforms(interpolation='NearestNeighbor'),
                           name='MaskToMNI')

    # Write corrected file in the designated output dir
    ds_mni = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='hmc_mni'), name='DerivativesHMCMNI')
    ds_mni_mask = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='hmc_mni_bmask'), name='DerivativesHMCMNImask')

    workflow.connect([
        (inputnode, pick_1st, [('epi_ras', 'in_file')]),
        (inputnode, ds_mni, [('epi', 'source_file')]),
        (inputnode, ds_mni_mask, [('epi', 'source_file')]),
        (pick_1st, gen_ref, [('roi_file', 'moving_image')]),
        (inputnode, merge_transforms, [('t1_2_mni_forward_transform', 'in1'),
                                       (('mat_epi_to_t1', _aslist), 'in2'),
                                       ('hmc_xforms', 'in3')]),
        (inputnode, mask_merge_tfms, [('t1_2_mni_forward_transform', 'in1'),
                                      (('mat_epi_to_t1', _aslist), 'in2')]),
        (inputnode, split, [('epi_ras', 'in_file')]),
        (split, epi_to_mni_transform, [('out_files', 'input_image')]),
        (merge_transforms, epi_to_mni_transform, [('out', 'transforms')]),
        (gen_ref, epi_to_mni_transform, [('out_file', 'reference_image')]),
        (epi_to_mni_transform, merge, [('output_image', 'in_files')]),
        (merge, ds_mni, [('merged_file', 'in_file')]),
        (mask_merge_tfms, mask_mni_tfm, [('out', 'transforms')]),
        (gen_ref, mask_mni_tfm, [('out_file', 'reference_image')]),
        (inputnode, mask_mni_tfm, [('epi_mask', 'input_image')]),
        (mask_mni_tfm, ds_mni_mask, [('output_image', 'in_file')])
    ])

    return workflow

# pylint: disable=R0914
def epi_unwarp(name='EPIUnwarpWorkflow', settings=None):
    """ A workflow to correct EPI images """
    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(
        fields=['epi', 'sbref_brain', 'fmap_fieldcoef', 'fmap_movpar',
                'fmap_mask', 'epi_brain']), name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['epi_correct', 'epi_mean']),
        name='outputnode'
    )

    # Read metadata
    meta = pe.MapNode(
        ReadSidecarJSON(fields=['TotalReadoutTime', 'PhaseEncodingDirection']),
        iterfield=['in_file'],
        name='metadata'
    )

    encfile = pe.Node(interface=niu.Function(
        input_names=['fieldmaps', 'in_dict'], output_names=['parameters_file'],
        function=create_encoding_file), name='TopUp_encfile', updatehash=True)

    fslsplit = pe.Node(fsl.Split(dimension='t'), name='EPI_split')

    # Now, we cannot use the LSR method
    unwarp_epi = pe.MapNode(fsl.ApplyTOPUP(method='jac', in_index=[1]),
                            iterfield=['in_files'], name='TopUpApply')

    # Merge back
    fslmerge = pe.Node(fsl.Merge(dimension='t'), name='EPI_corr_merge')

    # Compute mean
    epi_mean = pe.Node(fsl.MeanImage(dimension='T'), name='EPI_mean')

    workflow.connect([
        (inputnode, meta, [('epi', 'in_file')]),
        (inputnode, encfile, [('epi', 'fieldmaps')]),
        (inputnode, fslsplit, [('epi_brain', 'in_file')]),
        (meta, encfile, [('out_dict', 'in_dict')]),
        (inputnode, unwarp_epi, [('fmap_fieldcoef', 'in_topup_fieldcoef'),
                                 ('fmap_movpar', 'in_topup_movpar')]),
        (encfile, unwarp_epi, [('parameters_file', 'encoding_file')]),
        (fslsplit, unwarp_epi, [('out_files', 'in_files')]),
        (unwarp_epi, fslmerge, [('out_corrected', 'in_files')]),
        (fslmerge, epi_mean, [('merged_file', 'in_file')]),
        (fslmerge, outputnode, [('merged_file', 'epi_correct')]),
        (epi_mean, outputnode, [('out_file', 'epi_mean')])
    ])

    # Plot result
    png_epi_corr= pe.Node(niu.Function(
        input_names=['in_file', 'overlay_file', 'out_file'], output_names=['out_file'],
        function=stripped_brain_overlay), name='PNG_epi_corr')
    png_epi_corr.inputs.out_file = 'corrected_EPI.png'

    ds_png = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='sdc'), name='DerivativesPNG')

    workflow.connect([
        (epi_mean, png_epi_corr, [('out_file', 'overlay_file')]),
        (inputnode, png_epi_corr, [('fmap_mask', 'in_file')]),
        (inputnode, ds_png, [('epi', 'source_file')]),
        (png_epi_corr, ds_png, [('out_file', 'in_file')])
    ])

    return workflow


def _gen_reference(fixed_image, moving_image, out_file=None):
    import os.path as op
    import numpy as np
    import nibabel as nb
    from nibabel.affines import apply_affine

    if out_file is None:
        fname, ext = op.splitext(op.basename(fixed_image))
        if ext == '.gz':
            fname, ext2 = op.splitext(fname)
            ext = ext2 + ext
        out_file = op.abspath('%s_wm%s' % (fname, ext))

    imref = nb.load(fixed_image)
    immov = nb.load(moving_image)

    orig = apply_affine(imref.affine, [-0.5] * 3)
    end = apply_affine(imref.affine, [s - 0.5 for s in imref.get_data().shape[:3]])

    mov_spacing = immov.get_header().get_zooms()[:3]
    new_sizes = np.ceil((end-orig)/mov_spacing)

    new_affine = immov.affine
    ref_center = apply_affine(imref.affine, (0.5 * (np.array(imref.get_data().shape[:3]))))

    new_center = new_affine[:3, :3].dot(new_sizes)
    new_affine[:3, 3] = -0.5 * new_center + ref_center

    new_ref_im = nb.Nifti1Image(np.zeros(tuple(new_sizes.astype(int))),
                                new_affine, immov.get_header())
    new_ref_im.to_filename(out_file)

    return out_file



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Preprocessing workflows for :abbr:`SB (single-band)`-reference (SBRef)
images.

Originally coded by Craig Moodie. Refactored by the CRN Developers.
"""

import os
import os.path as op
from nipype.pipeline import engine as pe
from nipype.interfaces import io as nio
from nipype.interfaces import utility as niu
from nipype.interfaces import fsl
from nipype.interfaces.ants.segmentation import N4BiasFieldCorrection

from fmriprep.utils.misc import gen_list
from fmriprep.interfaces import ReadSidecarJSON
from fmriprep.workflows.fieldmap.se_pair_workflow import create_encoding_file
from fmriprep.viz import stripped_brain_overlay

def sbref_workflow(name='SBrefPreprocessing', settings=None):
    """SBref processing workflow"""
    if settings is None:
        settings = {}

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=['sbref', 'fmap_ref_brain', 'fmap_mask',
                        'fmap_fieldcoef', 'fmap_movpar']), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=['sbref_unwarped']), name='outputnode')

    # Read metadata
    meta = pe.MapNode(ReadSidecarJSON(fields=['TotalReadoutTime', 'PhaseEncodingDirection']),
                      iterfield=['in_file'], name='metadata')

    encfile = pe.Node(interface=niu.Function(
        input_names=['fieldmaps', 'in_dict'], output_names=['parameters_file'],
        function=create_encoding_file), name='TopUp_encfile', updatehash=True)

    #  Skull strip SBRef to get reference brain
    sbref_bet = pe.MapNode(fsl.BET(mask=True, functional=True, frac=0.6),
                           iterfield=['in_file'], name="sbref_bet")

    # Head motion correction (using fieldmap magnitude as reference)
    fslmerge = pe.Node(fsl.Merge(dimension='t'), name='SBref_merge')
    hmc_se = pe.Node(fsl.MCFLIRT(cost='normcorr', mean_vol=True), name='SBref_head_motion_corr')
    fslsplit = pe.Node(fsl.Split(dimension='t'), name='SBref_split')

    # Use the least-squares method to correct the dropout of the SBRef images
    unwarp_mag = pe.Node(fsl.ApplyTOPUP(method='lsr'), name='TopUpApply')

    # Remove bias
    inu_n4 = pe.Node(N4BiasFieldCorrection(dimension=3), name='SBref_bias')

    workflow.connect([
        (inputnode, meta, [('sbref', 'in_file')]),
        (inputnode, sbref_bet, [('sbref', 'in_file')]),
        (inputnode, encfile, [('sbref', 'fieldmaps')]),
        (inputnode, hmc_se, [('fmap_ref_brain', 'ref_file')]),
        (sbref_bet, fslmerge, [('out_file', 'in_files')]),
        (inputnode, unwarp_mag, [('fmap_fieldcoef', 'in_topup_fieldcoef'),
                                 ('fmap_movpar', 'in_topup_movpar')]),
        (meta, encfile, [('out_dict', 'in_dict')]),
        (fslmerge, hmc_se, [('merged_file', 'in_file')]),
        (encfile, unwarp_mag, [('parameters_file', 'encoding_file')]),
        (hmc_se, fslsplit, [('out_file', 'in_file')]),
        (fslsplit, unwarp_mag, [('out_files', 'in_files'),
                                (('out_files', gen_list), 'in_index')]),
        (unwarp_mag, inu_n4, [('out_corrected', 'input_image')]),
        (inu_n4, outputnode, [('output_image', 'sbref_unwarped')])
    ])

    # Plot result
    png_sbref_corr= pe.Node(niu.Function(
        input_names=["in_file", "overlay_file", "out_file"], output_names=["out_file"],
        function=stripped_brain_overlay), name="PNG_sbref_corr")
    png_sbref_corr.inputs.out_file = "corrected_SBRef.png"

    datasink = pe.Node(nio.DataSink(base_directory=op.join(settings['work_dir'], "images")),
                       name="datasink", parameterization=False)

    workflow.connect([
        (inu_n4, png_sbref_corr, [('output_image', 'overlay_file')]),
        (inputnode, png_sbref_corr, [('fmap_mask', 'in_file')]),
        (png_sbref_corr, datasink, [('out_file', '@corrected_SBRef')])
    ])
    return workflow

def sbref_t1_registration(name='SBrefSpatialNormalization', settings=None):
    """
    Uses FSL FLIRT with the BBR cost function to find the transform that
    maps the SBRef space into the T1-space
    """
    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=['sbref_brain', 't1_brain', 't1_seg']),
                        name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=['mat_sbr_to_t1', 'mat_t1_to_sbr']),
                         name='outputnode')

    # Extract wm mask from segmentation
    wm_mask = pe.Node(niu.Function(input_names=['in_file'], output_names=['out_file'],
                      function=_extract_wm), name='WM_mask')

    flt_bbr = pe.Node(fsl.FLIRT(dof=6, cost_func='bbr'), name="Flirt_BBR")
    flt_bbr.inputs.schedule = settings['fsl'].get(
        'flirt_bbr', op.join(os.getenv('FSLDIR'), 'etc/flirtsch/bbr.sch'))

    # make equivalent warp fields
    invt_bbr = pe.Node(fsl.ConvertXFM(invert_xfm=True), name="Flirt_BBR_Inv")

    workflow.connect([
        (inputnode, wm_mask, [('t1_seg', 'in_file')]),
        (inputnode, flt_bbr, [('sbref_brain', 'in_file'),
                              ('t1_brain', 'reference')]),
        (wm_mask, flt_bbr, [('out_file', 'wm_seg')]),
        (flt_bbr, invt_bbr, [('out_matrix_file', 'in_file')]),
        (flt_bbr, outputnode, [('out_matrix_file', 'mat_sbr_to_t1')]),
        (invt_bbr, outputnode, [('out_file', 'mat_t1_to_sbr')])
    ])

    # Plots for report
    png_sbref_t1= pe.Node(niu.Function(
        input_names=["in_file", "overlay_file", "out_file"], output_names=["out_file"],
        function=stripped_brain_overlay), name="PNG_sbref_t1")
    png_sbref_t1.inputs.out_file = "sbref_to_t1.png"

    datasink = pe.Node(nio.DataSink(base_directory=op.join(settings['work_dir'], "images")),
                       name="datasink", parameterization=False)

    workflow.connect([
        (flt_bbr, png_sbref_t1, [('out_file', 'overlay_file')]),
        (inputnode, png_sbref_t1, [('t1_seg', 'in_file')]),
        (png_sbref_t1, datasink, [('out_file', '@sbref_to_t1')])
    ])

    return workflow


def _extract_wm(in_file):
    import os.path as op
    import nibabel as nb
    import numpy as np

    im = nb.load(in_file)
    data = im.get_data().astype(np.uint8)
    data[data != 3] = 0
    data[data > 0] = 1

    out_file = op.abspath('wm_mask.nii.gz')
    nb.Nifti1Image(data, im.get_affine(), im.get_header()).to_filename(out_file)
    return out_file

###################################
# Deprecated code
###################################

def sbref_workflow_deprecated(name='SBrefPreprocessing', settings=None):
    """Legacy SBref processing workflow"""
    if settings is None:
        settings = {}

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(
        fields=['sbref', 'mag_brain', 'fmap_scaled', 'fmap_mask', 'fmap_unmasked',
                'in_topup', 't1', 't1_brain', 't1_seg']), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(
        fields=['sbref_unwarped', 'sbref_fmap', 'mag2sbref_matrix', 'sbref_brain',
                'sbref_brain_corrected', 't1_brain', 'wm_seg', 'sbref_2_t1_transform']), name='outputnode')

    #  Skull strip SBRef to get reference brain
    sbref_bet = pe.Node(
        fsl.BET(mask=True, functional=True, frac=0.6), name="sbref_bet")

    #  Unwarp SBRef using Fugue  (N.B. duplicated in epi_reg_workflow!!!!!)
    fugue_sbref = pe.Node(fsl.FUGUE(unwarp_direction='x', dwell_time=dwell_time),
                          name="SBRef_Unwarping")

    strip_corrected_sbref = pe.Node(fsl.BET(mask=True, frac=0.6, robust=True),
                                    name="BET_Corrected_SBRef")

    flt_sbref_brain_t1_brain = pe.Node(
        fsl.FLIRT(dof=6, bins=640, cost_func='mutualinfo'),
        name='sbref_brain_2_t1_brain_affine_transform'
    )

    flt_sbref_2_t1 = pe.Node(
        fsl.FLIRT(dof=6, bins=640, cost_func='mutualinfo'),
        name='sbref_2_t1_affine_transform'
    )

    # Affine transform of T1 segmentation into SBRref space
    flt_wmseg_sbref = pe.Node(fsl.FLIRT(dof=6, bins=64, cost_func='mutualinfo'),
                              name="WMSeg_2_SBRef_Brain_Affine_Transform")

    invert_wmseg_sbref = pe.Node(
        fsl.ConvertXFM(invert_xfm=True), name="invert_wmseg_sbref"
    )

    #  Run the commands from epi_reg_dof
    #  WITH FIELDMAP (unwarping steps)
    #  flirt -in ${fmapmagbrain} -ref ${vrefbrain} -dof ${dof} -omat
    #  ${vout}_fieldmap2str_init.mat
    flt_fmap_mag_brain_sbref_brain = pe.Node(
        fsl.FLIRT(dof=6, bins=640, cost_func='mutualinfo'),
        name="Fmap_Mag_Brain_2_SBRef_Brain_Affine_Transform"
    )

    #  flirt -in ${fmapmaghead} -ref ${vrefhead} -dof ${dof} -init
    #  ${vout}_fieldmap2str_init.mat -omat ${vout}_fieldmap2str.mat -out
    #  ${vout}_fieldmap2str -nosearch
    flt_fmap_mag_sbref = pe.Node(
        fsl.FLIRT(dof=6, no_search=True, bins=640, cost_func='mutualinfo'),
        name="Fmap_Mag_2_SBRef_Affine_Transform"
    )

    #  the following is a NEW HACK to fix extrapolation when fieldmap is too
    #  small
    #  applywarp -i ${vout}_fieldmaprads_unmasked -r ${vrefhead}
    #  --premat=${vout}_fieldmap2str.mat -o ${vout}_fieldmaprads2str_pad0
    aw_fmap_unmasked_sbref = pe.Node(fsl.ApplyWarp(relwarp=True),
                                     name="Apply_Warp_Fmap_Unmasked_2_SBRef")

    # fslmaths ${vout}_fieldmaprads2str_pad0 -abs -bin
    # ${vout}_fieldmaprads2str_innermask
    fmap_unmasked_abs = pe.Node(
        fsl.UnaryMaths(operation='abs'), name="Abs_Fmap_Unmasked_Warp")
    fmap_unmasked_bin = pe.Node(
        fsl.UnaryMaths(operation='bin'), name="Binarize_Fmap_Unmasked_Warp")

    # fugue --loadfmap=${vout}_fieldmaprads2str_pad0
    # --mask=${vout}_fieldmaprads2str_innermask --unmaskfmap
    # --unwarpdir=${fdir} --savefmap=${vout}_fieldmaprads2str_dilated
    fugue_dilate = pe.Node(
        fsl.FUGUE(unwarp_direction='x', dwell_time=dwell_time,
                  save_unmasked_fmap=True),
        name="Fmap_Dilating"
    )

    # epi_reg --epi=sub-S2529LVY1263171_task-nback_run-1_bold_brain
    # --t1=../Preprocessing_test_workflow/_subject_id_S2529LVY1263171/Bias_Field_Correction/sub-S2529LVY1263171_run-1_T1w_corrected.nii.gz
    # --t1brain=sub-S2529LVY1263171_run-1_T1w_corrected_bet_brain.nii.gz
    # --out=sub-S2529LVY1263171/func/sub-S2529LVY1263171_task-nback_run-1_bold_undistorted
    # --fmap=Topup_Fieldmap_rad.nii.gz
    # --fmapmag=fieldmap_topup_corrected.nii.gz
    # --fmapmagbrain=Magnitude_brain.nii.gz --echospacing=dwell_time
    # --pedir=x- -v

    sbref_stripped_overlay= pe.Node(
        niu.Function(
            input_names=["in_file", "overlay_file", "out_file"],
            output_names=["out_file"],
            function=stripped_brain_overlay
        ),
        name="sbref_stripped_overlay"
    )
    sbref_stripped_overlay.inputs.out_file = "sbref_stripped_overlay.png"

    datasink = pe.Node(
        interface=nio.DataSink(base_directory=op.join(settings['work_dir'], "images")),
        name="datasink",
        parameterization=False
    )

    workflow.connect([

        (inputnode, flt_wmseg_sbref, [('sbref', 'reference')]),
        (inputnode, flt_wmseg_sbref, [('t1_seg', 'in_file')]),
        (flt_wmseg_sbref, invert_wmseg_sbref, [('out_matrix_file', 'in_file')]),
        (flt_wmseg_sbref, outputnode, [('out_file', 'wm_seg')]),
        (invert_wmseg_sbref, outputnode, [('out_file', 'sbref_2_t1_transform')]),




        (inputnode, sbref_bet, [('sbref', 'in_file')]),
        (inputnode, fugue_sbref, [
            ('sbref', 'in_file'),
            ('fmap_scaled', 'fmap_in_file')
        ]),
        (inputnode, flt_fmap_mag_brain_sbref_brain, [('mag_brain', 'in_file')]),
        (inputnode, fugue_sbref, [('fmap_mask', 'mask_file')]),
        (inputnode, flt_fmap_mag_sbref, [('in_topup', 'in_file')]),
        (inputnode, aw_fmap_unmasked_sbref, [('fmap_unmasked', 'in_file')]),
        #  (inputnode, flt_sbref_brain_t1_brain, [('stripped_t1', 'reference')]),
        #  (strip_corrected_sbref, flt_sbref_brain_t1_brain, [('out_file', 'in_file')]),
        #  (fugue_sbref, flt_sbref_2_t1, [('unwarped_file', 'in_file')]),
        #  (inputnode, flt_sbref_2_t1, [('t1', 'reference')]),
        #  (flt_sbref_brain_t1_brain, flt_sbref_2_t1, [('out_matrix_file', 'in_matrix_file')]),
        #  (inputnode, SBRef_skull_strip, [("sbref", "in_file")]),
        (fugue_sbref, strip_corrected_sbref, [('unwarped_file', 'in_file')]),
        # might need to switch to [strip_corrected_sbref, "in_file"] here
        # instead of [sbref_bet, "out_file"]
        # might need to switch to [strip_corrected_sbref, "in_file"] here
        # instead of [sbref_bet, "out_file"]
        (sbref_bet, flt_fmap_mag_brain_sbref_brain, [('out_file', 'reference')]),
        (fugue_sbref, flt_fmap_mag_sbref, [('unwarped_file', 'reference')]),
        (flt_fmap_mag_brain_sbref_brain, flt_fmap_mag_sbref, [
            ('out_matrix_file', 'in_matrix_file')
        ]),
        (flt_fmap_mag_sbref, aw_fmap_unmasked_sbref, [('out_matrix_file', 'premat')]),
        (fugue_sbref, aw_fmap_unmasked_sbref, [('unwarped_file', 'ref_file')]),
        (aw_fmap_unmasked_sbref, fmap_unmasked_abs, [('out_file', 'in_file')]),
        (fmap_unmasked_abs, fmap_unmasked_bin, [('out_file', 'in_file')]),
        (aw_fmap_unmasked_sbref, fugue_dilate, [('out_file', 'fmap_in_file')]),
        (fmap_unmasked_bin, fugue_dilate, [('out_file', 'mask_file')]),
        (sbref_bet, outputnode, [('out_file', 'sbref_brain')]),
        (fugue_sbref, outputnode, [('unwarped_file', 'sbref_unwarped')]),
        (fugue_dilate, outputnode, [('fmap_out_file', 'sbref_fmap')]),
        (flt_fmap_mag_sbref, outputnode, [('out_matrix_file', 'mag2sbref_matrix')]),
        (strip_corrected_sbref, outputnode, [('out_file', 'sbref_brain_corrected')]),
        (inputnode, sbref_stripped_overlay, [('sbref', 'overlay_file')]),
        (sbref_bet, sbref_stripped_overlay, [('mask_file', 'in_file')]),
        (sbref_stripped_overlay, datasink, [('out_file', '@sbref_stripped_overlay')])
    ])
    return workflow



from future.utils import raise_from

from nipype.pipeline import Workflow

from fmriprep.workflows.fieldmap.fieldmap_to_phasediff import fieldmap_to_phasediff
from fmriprep.workflows.fieldmap.se_pair_workflow import se_pair_workflow
from fmriprep.workflows.fieldmap.phase_diff_and_magnitudes import phase_diff_and_magnitudes
from fmriprep.workflows.fieldmap.helper import (is_fmap_type)

def fieldmap_decider(subject_data, settings):
    ''' Initialize FieldmapDecider to automatically find a
    Fieldmap preprocessing workflow '''

    # POSSIBLE FILES ACCORDING TO BIDS 1.0.0
    # 8.9.1 one phase diff image, at least one magnitude image
    # 8.9.2 two phase images, two magnitude images
    # 8.9.3 fieldmap image (and one magnitude image)
    # 8.9.4 multiple phase encoded directions (topup)

    # inputs = { 'fieldmaps': None }
    # outputs = { 'fieldmaps': None,
    #             'outputnode.mag_brain': None,
    #             'outputnode.fmap_mask': None,
    #             'outputnode.fmap_fieldcoef': None,
    #             'outputnode.fmap_movpar': None}

    try:
        subject_data['fieldmaps'][0]
    except IndexError as e:
        raise_from(NotImplementedError("No fieldmap data found"), e)

    for filename in subject_data['fieldmaps']:
        if is_fmap_type('phasediff', filename): # 8.9.1
            return phase_diff_and_magnitudes()
        elif is_fmap_type('phase', filename): # 8.9.2
            raise NotImplementedError("No workflow for phase fieldmap data")
        elif is_fmap_type('fieldmap', filename): # 8.9.3
            return fieldmap_to_phasediff(settings=settings)
        elif is_fmap_type('topup', filename): # 8.0.4
            return se_pair_workflow(settings=settings)

    raise IOError("Unrecognized fieldmap structure")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Anatomical Reference -processing workflows.

Originally coded by Craig Moodie. Refactored by the CRN Developers.

"""
import os
import os.path as op
import pkg_resources as pkgr

from nipype.interfaces import ants
from nipype.interfaces import fsl
from nipype.interfaces import freesurfer as fs
from nipype.interfaces import io as nio
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe

from niworkflows.anat.skullstrip import afni_wf as skullstrip_wf
from niworkflows.common import reorient as mri_reorient_wf

from fmriprep.interfaces import DerivativesDataSink
from fmriprep.data import get_ants_oasis_template_ras, get_mni_template_ras
from fmriprep.viz import stripped_brain_overlay, anatomical_overlay


#  pylint: disable=R0914
def t1w_preprocessing(name='t1w_preprocessing', settings=None):
    """T1w images preprocessing pipeline"""

    if settings is None:
        raise RuntimeError('Workflow settings are missing')

    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(niu.IdentityInterface(fields=['t1']), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(
        fields=['t1_seg', 'bias_corrected_t1', 't1_brain', 't1_2_mni',
                't1_2_mni_forward_transform', 't1_2_mni_reverse_transform',
                't1_segmentation']), name='outputnode')

    # 1. Reorient T1
    arw = pe.Node(fs.MRIConvert(out_type='niigz', out_orientation='RAS'), name='Reorient')

    # 2. T1 Bias Field Correction
    inu_n4 = pe.Node(ants.N4BiasFieldCorrection(dimension=3), name='Bias_Field_Correction')

    # 3. Skull-stripping
    asw = skullstrip_wf()
    if settings.get('skull_strip_ants', False):
        asw = skullstrip_ants(settings=settings)

    # 4. Segmentation
    t1_seg = pe.Node(fsl.FAST(no_bias=True), name='T1_Segmentation')

    # 5. T1w to MNI registration
    t1_2_mni = pe.Node(ants.Registration(), name='T1_2_MNI_Registration')
    t1_2_mni.inputs.collapse_output_transforms = False
    t1_2_mni.inputs.write_composite_transform = False
    t1_2_mni.inputs.output_transform_prefix = 'T1_to_MNI_'
    t1_2_mni.inputs.output_warped_image = 't1_to_mni.nii.gz'
    t1_2_mni.inputs.num_threads = 4
    t1_2_mni.inputs.fixed_image = op.join(get_mni_template_ras(), 'MNI152_T1_1mm.nii.gz')
    t1_2_mni.inputs.fixed_image_mask = op.join(
        get_mni_template_ras(), 'MNI152_T1_1mm_brain_mask.nii.gz')

    # Hack to avoid re-running ANTs all the times
    grabber_interface = nio.JSONFileGrabber()
    setattr(grabber_interface, '_always_run', False)
    t1_2_mni_params = pe.Node(grabber_interface, name='t1_2_mni_params')
    t1_2_mni_params.inputs.in_file = (
        pkgr.resource_filename('fmriprep', 'data/{}.json'.format(
            settings.get('ants_t1-mni_settings', 't1-mni_registration2')))
    )

    workflow.connect([
        (inputnode, arw, [('t1', 'in_file')]),
        (arw, inu_n4, [('out_file', 'input_image')]),
        (inu_n4, asw, [('output_image', 'inputnode.in_file')]),
        (asw, t1_seg, [('outputnode.out_file', 'in_files')]),
        (inu_n4, t1_2_mni, [('output_image', 'moving_image')]),
        (asw, t1_2_mni, [('outputnode.out_mask', 'moving_image_mask')]),
        (t1_seg, outputnode, [('tissue_class_map', 't1_seg')]),
        (inu_n4, outputnode, [('output_image', 'bias_corrected_t1')]),
        (t1_seg, outputnode, [('tissue_class_map', 't1_segmentation')]),
        (t1_2_mni, outputnode, [
            ('warped_image', 't1_2_mni'),
            ('forward_transforms', 't1_2_mni_forward_transform'),
            ('reverse_transforms', 't1_2_mni_reverse_transform')
        ]),
        (asw, outputnode, [
            ('outputnode.out_file', 't1_brain')]),
    ])

    # Connect reporting nodes
    t1_stripped_overlay = pe.Node(niu.Function(
        input_names=['in_file', 'overlay_file', 'out_file'], output_names=['out_file'],
        function=stripped_brain_overlay), name='PNG_T1_SkullStrip')
    t1_stripped_overlay.inputs.out_file = 't1_stripped_overlay.png'

    # The T1-to-MNI will be plotted using the segmentation. That's why we transform it first
    seg_2_mni = pe.Node(ants.ApplyTransforms(
        dimension=3, default_value=0, interpolation='NearestNeighbor'), name='T1_2_MNI_warp')
    seg_2_mni.inputs.reference_image = op.join(get_mni_template_ras(), 'MNI152_T1_1mm.nii.gz')

    t1_2_mni_overlay = pe.Node(niu.Function(
        input_names=['in_file', 'overlay_file', 'out_file'], output_names=['out_file'],
        function=stripped_brain_overlay), name='PNG_T1_to_MNI')
    t1_2_mni_overlay.inputs.out_file = 't1_to_mni_overlay.png'
    t1_2_mni_overlay.inputs.overlay_file = op.join(get_mni_template_ras(), 'MNI152_T1_1mm.nii.gz')

    datasink = pe.Node(
        interface=nio.DataSink(
            base_directory=op.join(settings['output_dir'], 'images')),
        name='datasink',
        parameterization=False
    )

    workflow.connect([
        (inu_n4, t1_stripped_overlay, [('output_image', 'overlay_file')]),
        (asw, t1_stripped_overlay, [('outputnode.out_mask', 'in_file')]),
        (t1_stripped_overlay, datasink, [('out_file', '@t1_stripped_overlay')]),
        (t1_seg, seg_2_mni, [('tissue_class_map', 'input_image')]),
        (t1_2_mni, seg_2_mni, [('forward_transforms', 'transforms'),
                               ('forward_invert_flags', 'invert_transform_flags')]),
        (seg_2_mni, t1_2_mni_overlay, [('output_image', 'in_file')]),
        (t1_2_mni_overlay, datasink, [('out_file', '@t1_2_mni_overlay')]),
    ])

    # Write corrected file in the designated output dir
    ds_t1_bias = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='inu'), name='DerivT1_inu')
    ds_t1_seg = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='inu_seg'), name='DerivT1_seg')
    ds_mask = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='bmask'), name='DerivT1_mask')

    ds_t1_mni = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='mni'), name='DerivT1w_MNI')
    ds_t1_mni_aff = pe.Node(
        DerivativesDataSink(base_directory=settings['output_dir'],
            suffix='mni_affine'), name='DerivT1w_MNI_affine')

    if settings.get('debug', False):
        workflow.connect([
            (t1_2_mni, ds_t1_mni_aff, [('forward_transforms', 'in_file')])
        ])
    else:
        ds_t1_mni_warp = pe.Node(
            DerivativesDataSink(base_directory=settings['output_dir'],
                suffix='mni_warp'), name='DerivT1w_MNI_warp')

        def _get_aff(inlist):
            return inlist[:-1]

        def _get_warp(inlist):
            return inlist[-1]

        workflow.connect([
            (inputnode, ds_t1_mni_warp, [('t1', 'source_file')]),
            (t1_2_mni, ds_t1_mni_aff, [
                (('forward_transforms', _get_aff), 'in_file')]),
            (t1_2_mni, ds_t1_mni_warp, [
                (('forward_transforms', _get_warp), 'in_file')])
        ])

    workflow.connect([
        (inputnode, ds_t1_bias, [('t1', 'source_file')]),
        (inputnode, ds_t1_seg, [('t1', 'source_file')]),
        (inputnode, ds_mask, [('t1', 'source_file')]),
        (inputnode, ds_t1_mni, [('t1', 'source_file')]),
        (inputnode, ds_t1_mni_aff, [('t1', 'source_file')]),
        (asw, ds_t1_bias, [('outputnode.out_file', 'in_file')]),
        (t1_seg, ds_t1_seg, [('tissue_class_map', 'in_file')]),
        (asw, ds_mask, [('outputnode.out_mask', 'in_file')]),
        (t1_2_mni, ds_t1_mni, [('warped_image', 'in_file')])
    ])

    # ANTs inputs connected here for clarity
    workflow.connect([
        (t1_2_mni_params, t1_2_mni, [
            ('metric', 'metric'),
            ('metric_weight', 'metric_weight'),
            ('dimension', 'dimension'),
            ('write_composite_transform', 'write_composite_transform'),
            ('radius_or_number_of_bins', 'radius_or_number_of_bins'),
            ('shrink_factors', 'shrink_factors'),
            ('smoothing_sigmas', 'smoothing_sigmas'),
            ('sigma_units', 'sigma_units'),
            ('output_transform_prefix', 'output_transform_prefix'),
            ('transforms', 'transforms'),
            ('transform_parameters', 'transform_parameters'),
            ('initial_moving_transform_com', 'initial_moving_transform_com'),
            ('number_of_iterations', 'number_of_iterations'),
            ('convergence_threshold', 'convergence_threshold'),
            ('convergence_window_size', 'convergence_window_size'),
            ('sampling_strategy', 'sampling_strategy'),
            ('sampling_percentage', 'sampling_percentage'),
            ('output_warped_image', 'output_warped_image'),
            ('use_histogram_matching', 'use_histogram_matching'),
            ('use_estimate_learning_rate_once',
             'use_estimate_learning_rate_once'),
            ('collapse_output_transforms', 'collapse_output_transforms'),
            ('winsorize_lower_quantile', 'winsorize_lower_quantile'),
            ('winsorize_upper_quantile', 'winsorize_upper_quantile'),
        ])
    ])

    return workflow


def skullstrip_ants(name='ANTsBrainExtraction', settings=None):
    if settings is None:
        settings = {'debug': False}

    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(niu.IdentityInterface(fields=['in_file']), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(
        fields=['out_file', 'out_mask']), name='outputnode')


    t1_skull_strip = pe.Node(ants.segmentation.BrainExtraction(
        dimension=3, use_floatingpoint_precision=1,
        debug=settings['debug']), name='Ants_T1_Brain_Extraction')
    t1_skull_strip.inputs.brain_template = op.join(get_ants_oasis_template_ras(),
                                                   'T_template0.nii.gz')
    t1_skull_strip.inputs.brain_probability_mask = op.join(
        get_ants_oasis_template_ras(),
        'T_template0_BrainCerebellumProbabilityMask.nii.gz'
    )
    t1_skull_strip.inputs.extraction_registration_mask = op.join(
        get_ants_oasis_template_ras(),
        'T_template0_BrainCerebellumRegistrationMask.nii.gz'
    )

    workflow.connect([
        (inputnode, t1_skull_strip, [('in_file', 'anatomical_image')]),
        (t1_skull_strip, outputnode, [('BrainExtractionMask', 'out_mask'),
                                      ('BrainExtractionBrain', 'out_file')])
    ])

    return workflow



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: oesteban
# @Date:   2015-11-19 16:44:27
# @Last Modified by:   oesteban
# @Last Modified time: 2016-06-03 12:14:53
"""
fMRI preprocessing workflow
=====
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from argparse import ArgumentParser
from argparse import RawTextHelpFormatter
from multiprocessing import cpu_count
import logging
import os
import os.path as op
from lockfile import LockFile


def main():
    """Entry point"""
    from nipype import config as ncfg
    from nipype.pipeline import engine as pe
    from fmriprep import __version__
    from fmriprep.workflows.base import fmri_preprocess_single
    from fmriprep.utils.misc import get_subject

    parser = ArgumentParser(description='fMRI Preprocessing workflow',
                            formatter_class=RawTextHelpFormatter)

    g_input = parser.add_argument_group('Inputs')
    g_input.add_argument('-B', '--bids-root', action='store', default=os.getcwd())
    g_input.add_argument('-S', '--subject-id', action='store', required=True)
    g_input.add_argument('-s', '--session-id', action='store', default='single_session')
    g_input.add_argument('-r', '--run-id', action='store', default='single_run')
    g_input.add_argument('-d', '--data-type', action='store', choices=['anat', 'func'])
    g_input.add_argument('-v', '--version', action='store_true', default=False,
                         help='Show current fmriprep version')
    g_input.add_argument('--debug', action='store_true', default=False,
                         help='run debug version of workflow')
    g_input.add_argument('--skull-strip-ants', action='store_true', default=False,
                         help='run debug version of workflow')

    g_input.add_argument('--nthreads', action='store', default=0,
                         type=int, help='number of threads')
    g_input.add_argument(
        "--write-graph", action='store_true', default=False,
        help="Write workflow graph.")
    g_input.add_argument(
        "--use-plugin", action='store', default=None,
        help='nipype plugin configuration file')

    g_outputs = parser.add_argument_group('Outputs')
    g_outputs.add_argument('-o', '--output-dir', action='store',
                           default=op.join(os.getcwd(), 'out'))
    g_outputs.add_argument('-w', '--work-dir', action='store',
                           default=op.join(os.getcwd(), 'work'))

    opts = parser.parse_args()

    if opts.version:
        print('fmriprep version ' + __version__)
        exit(0)

    # Warn for default work/output directories
    if (opts.work_dir == parser.get_default('work_dir') or
          opts.output_dir == parser.get_default('output_dir')):
        logging.warning("work-dir and/or output-dir not specified. Using " +
                        opts.work_dir + " and " + opts.output_dir)

    settings = {
        'bids_root': op.abspath(opts.bids_root),
        'write_graph': opts.write_graph,
        'nthreads': opts.nthreads,
        'debug': opts.debug,
        'skull_strip_ants': opts.skull_strip_ants,
        'output_dir': op.abspath(opts.output_dir),
        'work_dir': op.abspath(opts.work_dir)
    }

    if opts.debug:
        settings['ants_t1-mni_settings'] = 't1-mni_registration_test'

    log_dir = op.join(settings['work_dir'], 'log')

    # Check and create output and working directories
    # Using locks to prevent https://github.com/poldracklab/mriqc/issues/111
    with LockFile('.fmriprep-folders-lock'):
        if not op.exists(settings['output_dir']):
            os.makedirs(settings['output_dir'])

        if not op.exists(settings['work_dir']):
            os.makedirs(settings['work_dir'])

        if not op.exists(log_dir):
            os.makedirs(log_dir)

    # Set nipype config
    ncfg.update_config({
        'logging': {'log_directory': log_dir, 'log_to_file': True},
        'execution': {'crashdump_dir': log_dir}
    })

    # nipype plugin configuration
    plugin_settings = {'plugin': 'Linear'}
    if opts.use_plugin is not None:
        from yaml import load as loadyml
        with open(opts.use_plugin) as f:
            plugin_settings = loadyml(f)
    else:
        # Setup multiprocessing
        if settings['nthreads'] == 0:
            settings['nthreads'] = cpu_count()

        if settings['nthreads'] > 1:
            plugin_settings['plugin'] = 'MultiProc'
            plugin_settings['plugin_args'] = {'n_procs': settings['nthreads']}

    # Retrieve BIDS data
    imaging_data = get_subject(settings['bids_root'], opts.subject_id)

    # Build main workflow and run
    preproc_wf = fmri_preprocess_single(imaging_data, settings=settings)
    preproc_wf.base_dir = settings['work_dir']
    preproc_wf.run(**plugin_settings)

    if opts.write_graph:
        workflow.write_graph()
    #  workflow.run(**plugin_settings)



if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This pipeline is developed by the Poldrack lab at Stanford University
(https://poldracklab.stanford.edu/) for use at
the Center for Reproducible Neuroscience (http://reproducibility.stanford.edu/),
as well as for open-source software distribution.
"""

__version__ = '0.1.1a1'


__author__ = 'The CRN developers'
__copyright__ = 'Copyright 2016, Center for Reproducible Neuroscience, Stanford University'
__credits__ = ['Craig Moodie', 'Ross Blair', 'Oscar Esteban', 'Chris F. Gorgolewski',
               'Russell A. Poldrack']
__license__ = '3-clause BSD'
__maintainer__ = 'Ross Blair'
__email__ = 'crn.poldracklab@gmail.com'
__status__ = 'Prototype'
__url__ = 'https://github.com/poldracklab/preprocessing-workflow'
__packagename__ = 'fmriprep'

__description__ = """Fmriprep is a functional magnetic resonance image pre-processing pipeline that
is designed to provide an easily accessible, state-of-the-art interface that is robust to differences
in scan acquisition protocols and that requires minimal user input, while providing easily interpretable
and comprehensive error and output reporting."""
__longdesc__ = """
This package is a functional magnetic resonance image preprocessing pipeline that is designed to
provide an easily accessible, state-of-the-art interface that is robust to differences in scan
acquisition protocols and that requires minimal user input, while providing easily interpretable
and comprehensive error and output reporting. This open-source neuroimaging data processing tool
is being developed as a part of the MRI image analysis and reproducibility platform offered by the
CRN. This pipeline is heavily influenced by the `Human Connectome Project analysis pipelines
<https://github.com/Washington-University/Pipelines>`_ and, as such, the backbone of this pipeline
is a python reimplementation of the HCP `GenericfMRIVolumeProcessingPipeline.sh` script. However, a
major difference is that this pipeline is executed using a `nipype workflow framework
<http://nipype.readthedocs.io/en/latest/>`_. This allows for each call to a software module or binary
to be controlled within the workflows, which removes the need for manual curation at every stage, while
still providing all the output and error information that would be necessary for debugging and interpretation
purposes. The fmriprep pipeline primarily utilizes FSL tools, but also utilizes ANTs tools at several stages
such as skull stripping and template registration. This pipeline was designed to provide the best software
implementation for each state of preprocessing, and will be updated as newer and better neuroimaging software
become available.
"""



from fmriprep.utils import misc
import re

def sort_fmaps(fieldmaps): # i.e. filenames
    from fmriprep.utils import misc
    from fmriprep.workflows.fieldmap.helper import is_fmap_type

    fmaps = {}
    for fmap_type in misc.fieldmap_suffixes.keys():
        fmaps[fmap_type] = []
        fmaps[fmap_type] = [doc for doc in fieldmaps
                            if is_fmap_type(fmap_type, doc)]
    return fmaps
        
def is_fmap_type(fmap_type, filename):
    return re.search(misc.fieldmap_suffixes[fmap_type], filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Data fetchers module """

from fmriprep.data.getters import (get_mni_template,
	                               get_mni_template_ras,
	                               get_ants_oasis_template,
                                   get_ants_oasis_template_ras)



from __future__ import division

import logging

from nipype.interfaces.io import JSONFileGrabber
from nipype.interfaces import utility as niu
from nipype.interfaces import ants
from nipype.interfaces import fsl
from nipype.pipeline import engine as pe
from nipype.workflows.dmri.fsl.utils import (siemens2rads, rads2radsec,
                                             demean_image, cleanup_edge_pipeline,
                                             add_empty_vol)

from fmriprep.utils.misc import fieldmap_suffixes

''' Fieldmap preprocessing workflow for fieldmap data structure
8.9.1 in BIDS 1.0.0: one phase diff and at least one magnitude image'''

def _sort_fmaps(fieldmaps):
    ''' just a little data massaging'''
    from fmriprep.workflows.fieldmap.helper import sort_fmaps

    fmaps = sort_fmaps(fieldmaps)
    # there is only one phasediff image
    # there may be more than one magnitude image, but the workflow (GetFirst
    # node) needs to change first
    return fmaps['phasediff'][0], fmaps['magnitude'][0]

# based on
# https://github.com/nipy/nipype/blob/bd36a5dadab73e39d8d46b1f1ad826df3fece5c1/nipype/workflows/dmri/fsl/artifacts.py#L514
def phase_diff_and_magnitudes(name='phase_diff_and_magnitudes', interp='Linear',
                              fugue_params=dict(smooth3d=2.0)):
    """
    Phase
    unwrapping is performed using `PRELUDE
    <http://fsl.fmrib.ox.ac.uk/fsl/fsl-4.1.9/fugue/prelude.html>`_
    [Jenkinson03]_. Preparation of the fieldmap is performed reproducing the
    script in FSL `fsl_prepare_fieldmap
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FUGUE/Guide#SIEMENS_data>`_.

    .. warning:: Only SIEMENS format fieldmaps are supported.

    .. admonition:: References

    .. [Jenkinson03] Jenkinson M., `Fast, automated, N-dimensional
    phase-unwrapping algorithm <http://dx.doi.org/10.1002/mrm.10354>`_,
    MRM 49(1):193-197, 2003, doi: 10.1002/mrm.10354.

    """
    inputnode = pe.Node(niu.IdentityInterface(fields=['fieldmaps']),
                        name='inputnode')

    outputnode = pe.Node(niu.IdentityInterface(fields=['mag_brain',
                                                       'fmap_mask',
                                                       'fmap_fieldcoef', # in sepair this comes from topup.out_fieldcoef; ask what to do
                                                       'fmap_movpar']), # same as above; topup.out_movpar
                         name='outputnode')

    vsm2topup = pe.Node(niu.Function(function=_vsm_to_topup,
                                     input_names=['vsm'], # voxel shift map file
                                     output_names=['fmap_fieldcoef', 'fmap_movpar']),
                        name='vsm2topup')
    sortfmaps = pe.Node(niu.Function(function=_sort_fmaps,
                                      input_names=['fieldmaps'],
                                      output_names=['phasediff', 'magnitude']),
                        name='SortFmaps')

    ingest_fmap_data = _ingest_fieldmap_data_workflow()

    r_params = _make_node_r_params()

    # oscar said it wasn't necessary but it appears to be
    eff_echo = pe.Node(niu.Function(function=_eff_t_echo,
                                    input_names=['echospacing', 'acc_factor'],
                                    output_names=['eff_echo']),
                       name='EffEcho')

    rad2rsec = pe.Node(niu.Function(
        input_names=['in_file', 'delta_te'], output_names=['out_file'],
        function=rads2radsec), name='ToRadSec')

    pre_fugue = pe.Node(fsl.FUGUE(save_fmap=True), name='PreliminaryFugue')

    wrangle_fmap_data = _wrangle_fieldmap_data_workflow()

    vsm = pe.Node(fsl.FUGUE(save_shift=True, **fugue_params),
                  name="ComputeVSM")

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, sortfmaps, [('fieldmaps', 'fieldmaps')]),
        (sortfmaps, ingest_fmap_data, [('phasediff', 'inputnode.bmap_pha'),
                                       ('magnitude', 'inputnode.bmap_mag')]),
        (ingest_fmap_data, vsm, [('outputnode.skull_strip_mask_file', 'mask_file')]),
        (ingest_fmap_data, outputnode, [('outputnode.mag_brain', 'mag_brain'), # ??? verify
                                         ('outputnode.skull_strip_mask_file', 'fmap_mask')]), # ??? verify
        (ingest_fmap_data, wrangle_fmap_data, [('outputnode.skull_strip_mask_file',
                                                'inputnode.in_mask')]), # ??? verify
        (ingest_fmap_data, rad2rsec, [('outputnode.unwrapped_phase_file', 'in_file')]),
        (rad2rsec, pre_fugue, [('out_file','fmap_in_file')]), # ??? verify

        (ingest_fmap_data, pre_fugue, [('outputnode.skull_strip_mask_file', # ??? verify
                                        'mask_file')]), # ??? verify
        (pre_fugue, wrangle_fmap_data, [('fmap_out_file',
                                         'inputnode.fmap_out_file')]),
        (wrangle_fmap_data, vsm, [('outputnode.out_file', 'fmap_in_file')]),

        (r_params, rad2rsec, [('delta_te', 'delta_te')]),
        (r_params, vsm, [('delta_te', 'asym_se_time')]),
        (r_params, eff_echo, [('echospacing', 'echospacing'),
                              ('acc_factor', 'acc_factor')]),
        (eff_echo, vsm, [('eff_echo', 'dwell_time')]),

        (vsm, vsm2topup, [('shift_out_file', 'vsm')]),
        (vsm2topup, outputnode, [('fmap_fieldcoef', 'fmap_fieldcoef'),
                                 ('fmap_movpar', 'fmap_movpar')])
    ])
    return wf

def _make_node_r_params():
    # find these values in data--see bids spec
    # delta_te from phase diff
    # echospacing, acc_factor, enc_dir from sbref/epi
    epi_defaults = {'delta_te': 2.46e-3, 'echospacing': 0.77e-3,
                    'acc_factor': 2, 'enc_dir': u'AP'}
    return pe.Node(JSONFileGrabber(defaults=epi_defaults),
                   name='SettingsGrabber')

def _ingest_fieldmap_data_workflow():
    ''' Takes phase diff and one magnitude file, handles the
    illumination problem, extracts the brain, and does phase
    unwrapping usig FSL's PRELUDE'''
    name="IngestFmapData"

    inputnode = pe.Node(niu.IdentityInterface(fields=['bmap_mag',
                                                      'bmap_pha']),
                        name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(
        fields=['unwrapped_phase_file',
                'skull_strip_mask_file',
                'mag_brain']),
                         name='outputnode')

    # ideally use mcflirt to align and then average w fsl something
    firstmag = pe.Node(fsl.ExtractROI(t_min=0, t_size=1), name='GetFirst')

    # de-gradient the fields ("illumination problem")
    n4 = pe.Node(ants.N4BiasFieldCorrection(dimension=3), name='Bias')

    bet = pe.Node(fsl.BET(frac=0.4, mask=True), name='BrainExtraction')
    # uses mask from bet; outputs a mask
    dilate = pe.Node(fsl.maths.MathsCommand(
        nan2zeros=True, args='-kernel sphere 5 -dilM'), name='MskDilate')

    # phase diff -> radians
    pha2rads = pe.Node(niu.Function(
        input_names=['in_file'], output_names=['out_file'],
        function=siemens2rads), name='PreparePhase')

    prelude = pe.Node(fsl.PRELUDE(process3d=True), name='PhaseUnwrap')

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, firstmag, [('bmap_mag', 'in_file')]),
        (firstmag, n4, [('roi_file', 'input_image')]),
        (n4, prelude, [('output_image', 'magnitude_file')]),
        (n4, bet, [('output_image', 'in_file')]),
        (bet, outputnode, [('mask_file', 'skull_strip_mask_file'), # ??? verify
                           ('out_file', 'mag_brain')]),

        (bet, dilate, [('mask_file', 'in_file')]),
        (dilate, prelude, [('out_file', 'mask_file')]),

        (inputnode, pha2rads, [('bmap_pha', 'in_file')]),
        (pha2rads, prelude, [('out_file', 'phase_file')]),
        (prelude, outputnode, [('unwrapped_phase_file',
                                'unwrapped_phase_file')]),
    ])
    return wf

def _wrangle_fieldmap_data_workflow():
    '''  Normalizes ("demeans") fieldmap data, cleans it up and
    organizes it for output'''
    name='WrangleFmapData'

    inputnode = pe.Node(niu.IdentityInterface(fields=['fmap_out_file',
                                                      'in_mask']),
                        name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=['out_file']),
                         name='outputnode')

    demean = pe.Node(niu.Function(
        input_names=['in_file', 'in_mask'], output_names=['out_file'],
        function=demean_image), name='DemeanFmap')

    cleanup = cleanup_edge_pipeline()

    addvol = pe.Node(niu.Function(
        input_names=['in_file'], output_names=['out_file'],
        function=add_empty_vol), name='AddEmptyVol')

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, demean, [('fmap_out_file', 'in_file'),
                             ('in_mask', 'in_mask')]),
        (demean, cleanup, [('out_file', 'inputnode.in_file')]),

        (inputnode, cleanup, [('in_mask', 'inputnode.in_mask')]),
        (cleanup, addvol, [('outputnode.out_file', 'in_file')]),
        (addvol, outputnode, [('out_file', 'out_file')]),
    ])
    return wf

def _vsm_to_topup(vsm):
    # OSCAR'S CODE HERE
    raise NotImplementedError()
    return fmap_fieldcoef, fmap_movpar

def _eff_t_echo(echospacing, acc_factor):
    eff_echo = echospacing / (1.0 * acc_factor)
    return eff_echo



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# @Author: oesteban
# @Date:   2016-06-03 09:35:13
# @Last Modified by:   oesteban
# @Last Modified time: 2016-06-03 10:06:55
import os
import os.path as op
from shutil import copy
import re
import simplejson as json
from nipype.interfaces.base import (traits, isdefined, TraitedSpec, BaseInterface,
                                    BaseInterfaceInputSpec, File, InputMultiPath,
                                    OutputMultiPath)
from lockfile import LockFile


class DerivativesDataSinkInputSpec(BaseInterfaceInputSpec):
    base_directory = traits.Directory(
        desc='Path to the base directory for storing data.')
    in_file = InputMultiPath(File(), exists=True, mandatory=True,
                             desc='the object to be saved')
    source_file = File(exists=True, mandatory=True, desc='the input func file')
    suffix = traits.Str('', mandatory=True, desc='suffix appended to source_file')

class DerivativesDataSinkOutputSpec(TraitedSpec):
    out_file = OutputMultiPath(File(exists=True, desc='written file path'))

class DerivativesDataSink(BaseInterface):
    input_spec = DerivativesDataSinkInputSpec
    output_spec = DerivativesDataSinkOutputSpec
    _always_run = True

    def __init__(self, **inputs):
        self._results = {'out_file': []}
        super(DerivativesDataSink, self).__init__(**inputs)

    def _run_interface(self, runtime):
        fname, _ = _splitext(self.inputs.source_file)
        _, ext = _splitext(self.inputs.in_file[0])

        m = re.search(
            '^(?P<subject_id>sub-[a-zA-Z0-9]+)(_(?P<ses_id>ses-[a-zA-Z0-9]+))?'
            '(_(?P<task_id>task-[a-zA-Z0-9]+))?(_(?P<acq_id>acq-[a-zA-Z0-9]+))?'
            '(_(?P<rec_id>rec-[a-zA-Z0-9]+))?(_(?P<run_id>run-[a-zA-Z0-9]+))?',
            fname
        )

        # TODO this quick and dirty modality detection needs to be implemented
        # correctly
        mod = 'func'
        if 'anat' in op.dirname(self.inputs.source_file):
            mod = 'anat'
        elif 'dwi' in op.dirname(self.inputs.source_file):
            mod = 'dwi'

        base_directory = os.getcwd()
        if isdefined(self.inputs.base_directory):
            base_directory = op.abspath(self.inputs.base_directory)

        out_path = 'derivatives/{subject_id}'.format(**m.groupdict())
        if m.groupdict().get('ses_id') is not None:
            out_path += '/{ses_id}'.format(**m.groupdict())
        out_path += '/{}'.format(mod)

        out_path = op.join(base_directory, out_path)
        with LockFile(op.join(base_directory, '.fmriprep-lock')):
            if not op.exists(out_path):
                os.makedirs(out_path)

        base_fname = op.join(out_path, fname)

        formatstr = '{bname}_{suffix}{ext}'
        if len(self.inputs.in_file) > 1:
            formatstr = '{bname}_{suffix}{i:04d}{ext}'


        for i, fname in enumerate(self.inputs.in_file):
            out_file = formatstr.format(
                bname=base_fname,
                suffix=self.inputs.suffix,
                i=i,
                ext=ext)
            self._results['out_file'].append(out_file)
            copy(self.inputs.in_file[i], out_file)

        return runtime

    def _list_outputs(self):
        return self._results


class ReadSidecarJSONInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc='the input nifti file')
    fields = traits.List(traits.Str, desc='get only certain fields')


class ReadSidecarJSONOutputSpec(TraitedSpec):
    out_dict = traits.Dict()


class ReadSidecarJSON(BaseInterface):
    """
    An utility to find and read JSON sidecar files of a BIDS tree
    """

    input_spec = ReadSidecarJSONInputSpec
    output_spec = ReadSidecarJSONOutputSpec

    def __init__(self, **inputs):
        self._results = {}
        super(ReadSidecarJSON, self).__init__(**inputs)

    def _run_interface(self, runtime):
        metadata = get_metadata_for_nifti(self.inputs.in_file)

        if isdefined(self.inputs.fields) and self.inputs.fields:
            for fname in self.inputs.fields:
                self._results[fname] = metadata[fname]
        else:
            self._results = metadata

        return runtime

    def _list_outputs(self):
        out = self.output_spec().get()
        out['out_dict'] = self._results
        return out

def get_metadata_for_nifti(in_file):
    """Fetchs metadata for a given nifi file"""
    in_file = op.abspath(in_file)

    fname, ext = op.splitext(in_file)
    if ext == '.gz':
        fname, ext2 = op.splitext(fname)
        ext = ext2 + ext

    side_json = fname + '.json'
    fname_comps = side_json.split('/')[-1].split("_")

    session_comp_list = []
    subject_comp_list = []
    top_comp_list = []
    ses = None
    sub = None

    for comp in fname_comps:
        if comp[:3] != "run":
            session_comp_list.append(comp)
            if comp[:3] == "ses":
                ses = comp
            else:
                subject_comp_list.append(comp)
                if comp[:3] == "sub":
                    sub = comp
                else:
                    top_comp_list.append(comp)


    top_json = "/" + "_".join(top_comp_list)
    potential_json = [top_json]

    subject_json = "/" + sub + "/" + "_".join(subject_comp_list)
    potential_json.append(subject_json)

    if ses:
        session_json = "/" + sub + "/" + ses + "/" + "_".join(session_comp_list)
        potential_json.append(session_json)

    potential_json.append(side_json)

    merged_param_dict = {}
    for json_file_path in potential_json:
        if op.isfile(json_file_path):
            with open(json_file_path, 'r') as jsonfile:
                param_dict = json.load(jsonfile)
                merged_param_dict.update(param_dict)

    return merged_param_dict

def _splitext(fname):
    fname, ext = op.splitext(op.basename(fname))
    if ext == '.gz':
        fname, ext2 = op.splitext(fname)
        ext = ext2 + ext
    return fname, ext



# -*- coding: utf-8 -*-
"""
Created on Sun Jul 26 16:54:08 2015

@author: craigmoodie
"""


def generate_report(output_file, first_plot, second_plot, third_plot,
                    fourth_plot, fifth_plot, sixth_plot, seventh_plot,
                    eighth_plot, t1_2_mni_plot=None):
    import os.path
    import pylab as plt
    from matplotlib.gridspec import GridSpec
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.image as mimage

    plots = [first_plot, second_plot, third_plot, fourth_plot,
             fifth_plot, sixth_plot, seventh_plot, eighth_plot]

    if t1_2_mni_plot is not None:
        plots.insert(6, t1_2_mni_plot)

    #  error_image = ""
    #  error = mimage.imread(error_image)
    report = PdfPages(output_file)
    fig = plt.figure()
    grid = GridSpec(3, 3)

    plot_iterator = iter(plots)
    for i in range(0, 3):
        for j in range(0, 3):
            try:
                plot = next(plot_iterator)
            except:
                print('{}, {}'.format(i, j))
                continue
            if plot is None:
                #  plot = error
                continue
            img = mimage.imread(plot)
            ax = plt.subplot(grid[j, i])
            ax.imshow(img)
            ax.set_aspect('auto')
            for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                         ax.get_xticklabels() + ax.get_yticklabels()):
                item.set_fontsize(8)

            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)

    fig.subplots_adjust(wspace=0.02, hspace=.2)

    report.savefig(fig, dpi=300)
    report.close()

    return os.path.abspath(output_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: oesteban
# @Date:   2015-11-19 16:44:27
# @Last Modified by:   oesteban
# @Last Modified time: 2016-07-29 16:38:19
""" fmriprep setup script """
import os
import sys

from fmriprep import (__version__, __email__, __url__, __packagename__, __license__,
                      __description__, __longdesc__, __maintainer__, __author__)


REQ_LINKS = []
with open('requirements.txt', 'r') as rfile:
    REQUIREMENTS = [line.strip() for line in rfile.readlines()]

for i, req in enumerate(REQUIREMENTS):
    if req.startswith('-e'):
        REQUIREMENTS[i] = req.split('=')[1]
        REQ_LINKS.append(req.split()[1])

if REQUIREMENTS is None:
    REQUIREMENTS = []

def main():
    """ Install entry-point """
    from glob import glob
    from setuptools import setup

    setup(
        name=__packagename__,
        version=__version__,
        description=__description__,
        long_description=__longdesc__,
        author=__author__,
        author_email='cmoodie@stanford.edu',
        email=__email__,
        maintainer=__maintainer__,
        maintainer_email=__email__,
        url=__url__,
        download_url='https://pypi.python.org/packages/source/f/fmriprep/'
                     'fmriprep-%s.tar.gz' % __version__,
        license=__license__,
        entry_points={'console_scripts': ['fmriprep=fmriprep.run_workflow:main',]},
        packages=['fmriprep',
                  'fmriprep.data',
                  'fmriprep.interfaces',
                  'fmriprep.utils',
                  'fmriprep.viz',
                  'fmriprep.workflows',
                  'fmriprep.workflows.fieldmap',
        ],
        package_data={'fmriprep': ['data/*.json']},
        install_requires=REQUIREMENTS,
        dependency_links=REQ_LINKS,
        zip_safe=False,
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: MRI processing',
            'Topic :: Scientific/Engineering :: Biomedical Imaging',
            'License :: OSI Approved :: 3-clause BSD License',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.5'
        ],
    )

if __name__ == '__main__':
    local_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    os.chdir(local_path)
    sys.path.insert(0, local_path)

    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype.interfaces import fsl

WORKFLOW_NAME = 'Fieldmap2Phasediff'

def fieldmap_to_phasediff(name=WORKFLOW_NAME, settings=None):
    """Legacy workflow to create a phasediff map from a fieldmap, to be digested by FUGUE"""

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=['fieldmap', 'fmap_mask', 'unwarp_direction',
                                                      'dwell_time']), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=['fmap_rads', 'fmap_unmasked']),
                         name='outputnode')

    # Convert topup fieldmap to rad/s [ 1 Hz = 6.283 rad/s]
    fmap_scale = pe.Node(fsl.BinaryMaths(operation='mul', operand_value=6.283),
                         name='fmap_scale')

    # Compute a mask from the fieldmap (??)
    fmap_abs = pe.Node(fsl.UnaryMaths(operation='abs', args='-bin'), name='fmap_abs')
    fmap_mul = pe.Node(fsl.BinaryMaths(operation='mul'), name='fmap_mul_mask')

    # Compute an smoothed field without mask
    fugue_unmask = pe.Node(fsl.FUGUE(save_unmasked_fmap=True), name='fmap_unmask')

    workflow.connect([
        (inputnode, fmap_scale, [('fieldmap', 'in_file')]),
        (inputnode, fmap_mul, [('fmap_mask', 'operand_file')]),
        (inputnode, fugue_unmask, [('unwarp_direction', 'unwarp_direction'),
                                   ('dwell_time', 'dwell_time')]),
        (fmap_scale, fmap_abs, [('out_file', 'in_file')]),
        (fmap_abs, fmap_mul, [('out_file', 'in_file')]),
        (fmap_scale, fugue_unmask, [('out_file', 'fmap_in_file')]),
        (fmap_mul, fugue_unmask, [('out_file', 'mask_file')]),
        (fmap_scale, outputnode, [('out_file', 'fmap_rads')]),
        (fugue_unmask, outputnode, [('fmap_out_file', 'fmap_unmasked')])
    ])
    return workflow



from glob import glob
import copy
import json
import os
import pprint
import re

INPUTS_SPEC = {'fieldmaps': [], 'epi': [], 'sbref': [], 't1': ''}

def gen_list(inlist, base=1):
    return range(base, len(inlist) + base)

def _walk_dir_for_prefix(target_dir, prefix):
    return [x for x in next(os.walk(target_dir))[1]
            if x.startswith(prefix)]


def get_subject(bids_root, subject_id, session_id=None, run_id=None,
                include_types=None):
    """
    Returns the imaging_data structure for the subject subject_id.
    If session is None, then the BIDS structure is not multisession.
    If run_id is None, it is assumed that the session does not have several
    runs.
    """
    if include_types is None:
        # Please notice that dwi is not here
        include_types = ['func', 'anat', 'fmap']
    subject_data = collect_bids_data(bids_root, include_types=None)
    subject_data = subject_data['sub-' + subject_id]

    if session_id is None:
        subject_data = subject_data[list(subject_data.keys())[0]]
    else:
        raise NotImplementedError

    if run_id is not None:
        raise NotImplementedError

    return subject_data


# if no scan_subject or scan_session are defined return all bids data for a
# given bids directory. Otherwise just the data for a given subject or scan
# can be returned
def collect_bids_data(dataset, include_types=None, scan_subject='sub-',
                      scan_session='ses-'):
    imaging_data = {}
    if include_types is None:
        include_types = ['func', 'anat', 'fmap', 'dwi']

    subjects = _walk_dir_for_prefix(dataset, scan_subject)
    if len(subjects) == 0:
        raise GeneratorExit("No BIDS subjects found to examine.")

    for subject in subjects:
        if subject not in imaging_data:
            imaging_data[subject] = {}
        subj_dir = os.path.join(dataset, subject)

        sessions = _walk_dir_for_prefix(subj_dir, scan_session)

        for scan_type in include_types:
            # seems easier to consider the case of multi-session vs.
            # single session separately?
            if len(sessions) > 0:
                subject_sessions = [os.path.join(subject, x)
                                    for x in sessions]
            else:
                subject_sessions = [subject]

            for session in subject_sessions:
                if session not in imaging_data[subject]:
                    imaging_data[subject][session] = copy.deepcopy(INPUTS_SPEC)
                scan_files = glob(os.path.join(dataset, session, scan_type,
                                               '*'))
                for scan_file in scan_files:
                    filename = scan_file.split('/')[-1]
                    filename_parts = filename.split('_')
                    modality = filename_parts[-1]
                    if 'sbref.nii' in modality:
                        imaging_data[subject][session]['sbref'].append(scan_file)
                    elif is_fieldmap_file(modality):
                        imaging_data[subject][session]['fieldmaps'].append(scan_file)
                    elif 'T1w.nii' in modality:
                        imaging_data[subject][session]['t1'] = scan_file
                    # temporary conditional until runs and tasks are handled
                    # in the imaging data structure
                    elif 'bold.nii' in filename:
                        imaging_data[subject][session]['epi'].append(scan_file)
                    else:
                        pass

    return imaging_data

def is_fieldmap_file(string):
    is_fieldmap_file = False
    for suffix in fieldmap_suffixes.values():
        if re.search(suffix, string):
            is_fieldmap_file = True
    return is_fieldmap_file

fieldmap_suffixes = {
    'phasediff': r"phasediff\.nii",
    'magnitude': r"magnitude[0-9]*\.nii",
    'phase': r"phase[0-9]+\.nii",
    'fieldmap': r"fieldmap\.nii",
    'topup': r"epi\.nii"
}


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
Created on Wed Jan 27 14:32:13 2016

@author: craigmoodie
"""

import os.path as op

from nipype.interfaces.fsl import MeanImage
from nipype.interfaces.utility import Function
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from nipype.pipeline.engine import Workflow, Node

from fmriprep.viz.pdf_compose import generate_report



def anatomical_overlay(in_file, overlay_file, out_file):
    import os.path
    import matplotlib as mpl
    mpl.use('Agg')
    from nilearn.plotting import plot_anat as plot_anat
    mask_display = plot_anat(in_file)
    mask_display.add_edges(overlay_file)
    mask_display.dim = -1
    #  mask_display.add_contours(overlay_file)
    #  mask_display.add_overlay(overlay_file)
    mask_display.title(out_file, x=0.01, y=0.99, size=15, color=None,
                       bgcolor=None, alpha=1)
    mask_display.savefig(out_file)
    return os.path.abspath(out_file)


def parcel_overlay(in_file, overlay_file, out_file):
    import os.path
    import matplotlib as mpl
    mpl.use('Agg')
    from nilearn.plotting import plot_epi
    mask_display = plot_epi(in_file)
    mask_display.add_edges(overlay_file)
    #  mask_display.add_contours(overlay_file)
    #  mask_display.add_overlay(overlay_file)
    mask_display.title(out_file, x=0.01, y=0.99, size=15, color=None,
                       bgcolor=None, alpha=1)
    mask_display.savefig(out_file)
    return os.path.abspath(out_file)


def stripped_brain_overlay(in_file, overlay_file, out_file):
    import os.path
    import nibabel as nb
    import matplotlib as mpl
    mpl.use('Agg')
    from nilearn.plotting import plot_roi
    vmax = nb.load(in_file).get_data().reshape(-1).max()
    mask_display = plot_roi(
        in_file, overlay_file, output_file=out_file, title=out_file,
        display_mode="ortho", dim=-1, alpha=.3, vmax=vmax + 1)
    #  mask_display.bg_img(overlay_file)
    #  mask_display.title(out_file, x=0.01, y=0.99, size=15, color=None,
    #                     bgcolor=None, alpha=1)
    #  mask_display.display_mode = "yx"
    mask_display
    return os.path.abspath(out_file)


def generate_report_workflow():

    report_workflow = Workflow(name="report_workflow")

    inputnode = pe.Node(niu.IdentityInterface(
        fields=['fmap_mag', 'fmap_mag_brain', 'raw_epi', 'stripped_epi',
                'corrected_epi_mean', 'sbref', 'sbref_brain', 'sbref_brain',
                'sbref_t1', 'corrected_sbref', 't1', 't1_brain', 't1_2_mni',
                'parcels_t1', 'parcels_native', 'fieldmap', 't1_segmentation',
                't1_wm_seg']),
        name='inputnode'
    )
    outputnode = pe.Node(niu.IdentityInterface(fields=['wm_seg']),
                         name='outputnode')

    #  Condensing 4D Stacks into 3D mean
    fmap_mag_mean = Node(MeanImage(), name="Fieldmap_mean")
    fmap_mag_mean.inputs.output_type = "NIFTI_GZ"
    fmap_mag_mean.inputs.dimension = 'T'

    stripped_epi_mean = Node(MeanImage(), name="Stripped_EPI_mean")
    stripped_epi_mean.inputs.output_type = "NIFTI_GZ"
    stripped_epi_mean.inputs.dimension = 'T'

    #  This isn't working!!!
    raw_epi_mean = Node(MeanImage(), name="Raw_EPI_mean")
    raw_epi_mean.inputs.output_type = "NIFTI_GZ"
    raw_epi_mean.inputs.dimension = 'T'
    fmap_overlay = Node(
        Function(
            input_names=["in_file", "overlay_file", "out_file"],
            output_names=["out_file"],
            function=parcel_overlay
        ),
        name="Fieldmap_to_SBRef_Overlay"
    )
    fmap_overlay.inputs.out_file = "Fieldmap_to_SBRef_Overlay.png"

    fmap_mag_BET = Node(
        Function(
            input_names=["in_file", "overlay_file", "out_file"],
            output_names=["out_file"],
            function=stripped_brain_overlay
        ),
        name="Fieldmap_Mag_BET"
    )
    fmap_mag_BET.inputs.out_file = "Fieldmap_Mag_BET.png"

    EPI_BET_report = Node(
        Function(
            input_names=["in_file", "overlay_file", "out_file"],
            output_names=["out_file"],
            function=stripped_brain_overlay
        ),
        name="EPI_Skullstrip_Overlay"
    )
    EPI_BET_report.inputs.out_file = "EPI_Skullstrip_Overlay.png"

    sbref_unwarp_overlay = Node(
        Function(
            input_names=["in_file", "overlay_file", "out_file"],
            output_names=["out_file"],
            function=anatomical_overlay
        ),
        name="SBRef_Unwarping_Overlay"
    )
    sbref_unwarp_overlay.inputs.out_file = "SBRef_Unwarping_Overlay.png"

    epi_unwarp_overlay = Node(
        Function(
            input_names=["in_file", "overlay_file", "out_file"],
            output_names=["out_file"],
            function=anatomical_overlay
        ),
        name="EPI_Unwarping_Overlay"
    )
    epi_unwarp_overlay.inputs.out_file = "EPI_Unwarping_Overlay.png"

    SBRef_BET = Node(
        Function(
            input_names=["in_file", "overlay_file", "out_file"],
            output_names=["out_file"],
            function=stripped_brain_overlay
        ),
        name="SBRef_BET"
    )
    SBRef_BET.inputs.out_file = "SBRef_BET_Overlay.png"

    epi_2_sbref = Node(
        Function(
            input_names=["in_file", "overlay_file", "out_file"],
            output_names=["out_file"],
            function=anatomical_overlay
        ),
        name="Functional_to_SBRef_Overlay"
    )
    epi_2_sbref.inputs.out_file = "Functional_to_SBRef_Overlay.png"

    sbref_2_t1 = Node(
        Function(
            input_names=["in_file", "overlay_file", "out_file"],
            output_names=["out_file"],
            function=anatomical_overlay
        ),
        name="SBRef_to_T1_Overlay"
    )
    sbref_2_t1.inputs.out_file = "SBRef_to_T1_Overlay.png"

#    final_pdf = Node(
#        Function(
#            input_names=[
#                "output_file", "first_plot", "second_plot", "third_plot",
#                "fourth_plot", "fifth_plot", "sixth_plot", "seventh_plot",
#                "eighth_plot"
#            ],
#            output_names=["output_file"],
#            function=generate_report
#        ),
#        name="Final_pdf"
#    )
#    final_pdf.inputs.output_file = "Preprocessing_Quality_Report.pdf"


    report_workflow.connect([
        (inputnode, fmap_mag_mean, [("fmap_mag", "in_file")]),
        (inputnode, fmap_overlay, [("fieldmap", "in_file")]),
        (inputnode, fmap_overlay, [("sbref", "overlay_file")]),
        (inputnode, fmap_mag_BET, [("fmap_mag_brain", "in_file")]),
        (fmap_mag_mean, fmap_mag_BET, [("out_file", "overlay_file")]),
        (inputnode, raw_epi_mean, [("raw_epi", "in_file")]),
        (inputnode, stripped_epi_mean, [("stripped_epi", "in_file")]),
        (stripped_epi_mean, EPI_BET_report, [("out_file", "in_file")]),
        (raw_epi_mean, EPI_BET_report, [("out_file", "overlay_file")]),
        (inputnode, sbref_unwarp_overlay, [("corrected_sbref", "in_file")]),
        (inputnode, sbref_unwarp_overlay, [("sbref", "overlay_file")]),
        (inputnode, SBRef_BET, [("sbref_brain", "in_file")]),
        (inputnode, SBRef_BET, [("sbref", "overlay_file")]),
        #  (inputnode, parcels_2_EPI, [("parcels_native", "in_file")]),
        #  (inputnode, parcels_2_EPI, [("corrected_epi_mean", "overlay_file")]),
        #  (inputnode, parcels_2_T1, [("parcels_t1", "in_file")]),
        #  (inputnode, parcels_2_T1, [("t1", "overlay_file")]),
        #  (inputnode, parcels_2_sbref, [("parcels_native", "in_file")]),
        #  prob should use corrected sbref brain mask
        #  (inputnode, parcels_2_sbref, [("corrected_sbref", "overlay_file")]),
        (inputnode, epi_2_sbref, [("corrected_epi_mean", "in_file")]),
        (inputnode, epi_2_sbref, [("corrected_sbref", "overlay_file")]),
        (inputnode, sbref_2_t1, [("sbref", "in_file")]),
        (inputnode, sbref_2_t1, [("t1_wm_seg", "overlay_file")]),
        (inputnode, epi_unwarp_overlay, [("corrected_epi_mean", "in_file")]),
        (raw_epi_mean, epi_unwarp_overlay, [("out_file", "overlay_file")]),
        #  (inputnode, T1_2_MNI, [("t1_mni", "in_file")]),
        #  replace sbref to mni and epi to mni with sbref and epi unwarping,
        #  also replace epi to t1 with parcel to sbref
#        (fmap_overlay, final_pdf, [("out_file", "first_plot")]),
#        (EPI_BET_report, final_pdf, [("out_file", "second_plot")]),
#        (SBRef_BET, final_pdf, [("out_file", "third_plot")]),
#        (sbref_unwarp_overlay, final_pdf, [("out_file",  "fifth_plot")]),
#        (epi_unwarp_overlay, final_pdf, [("out_file",  "sixth_plot")]),
#        (epi_2_sbref, final_pdf, [("out_file",  "seventh_plot")]),
#        # (t1_2_mni, final_pdf, [("out_file", "t1_2_mni_plot")]),
#        (sbref_2_t1, final_pdf, [("out_file",  "eighth_plot")]),
#        #  (T1_2_MNI, final_pdf, [("out_file",  "ninth_plot")]),
#        #  (parcels_2_T1, final_pdf, [("out_file",  "tenth_plot")]),
#        #  (parcels_2_EPI, final_pdf, [("out_file",  "eleventh_plot")]),
#        #  (parcels_2_sbref, final_pdf, [("out_file",  "twelfth_plot")]),
    ])

    return report_workflow


if __name__ == '__main__':
    report_wf = generate_report_workflow()
    report_wf.write_graph()
